from abc import ABC

import numpy as np
import tensorflow as tf

from mlc.models.model import Model


class TfModel(Model, ABC):
    def load(self, path):
        self.model = tf.keras.models.load_model(path)

    def save(self, path):
        self.model.save(path)

    def predict_proba(self, x: np.ndarray) -> np.ndarray:
        """
        Only implemented for binary / multi-class-classification tasks.
        Returns the probability distribution over the classes C.
        (Save probabilities to self.prediction_probabilities)

        :param x: test data
        :return: probabilities for the classes (Shape N x C)
        """

        prediction_probabilities = self.model.predict(x)

        # If binary task returns only probability for the true class,
        # adapt it to return (N x 2)
        if prediction_probabilities.shape[1] == 1:
            prediction_probabilities = np.concatenate(
                (
                    1 - prediction_probabilities,
                    prediction_probabilities,
                ),
                1,
            )
        return prediction_probabilities



import torch
from icecream import install

torch.set_num_threads(1)
install()

from . import env  # noqa
from .data import *  # noqa
from .deep import *  # noqa
from .env import *  # noqa
from .metrics import *  # noqa
from .util import *  # noqa



from typing import Dict, List

import numpy as np
import numpy.typing as npt
from sklearn.model_selection import train_test_split

from mlc.constraints.relation_constraint import BaseRelationConstraint
from mlc.constraints.relation_constraint import Constant as Co
from mlc.constraints.relation_constraint import Feature as Fe
from mlc.constraints.relation_constraint import SafeDivision, Value
from mlc.datasets.dataset import (
    CsvDataSource,
    Dataset,
    DataSource,
    DefaultIndexSorter,
    DownloadFileDataSource,
    Splitter,
    Task,
)


class Ctu13Splitter(Splitter):
    def get_splits(self, dataset: Dataset) -> Dict[str, npt.NDArray[np.int_]]:
        _, y = dataset.get_x_y()
        i = np.arange(len(y))
        i_train = i[:143046]
        i_test = i[143046:]
        i_train, i_val = train_test_split(
            i_train,
            random_state=1319,
            shuffle=True,
            stratify=y[i_train],
            test_size=0.2,
        )
        return {"train": i_train, "val": i_val, "test": i_test}


def get_relation_constraints(
    metadata: DataSource,
) -> List[BaseRelationConstraint]:

    features = metadata.load_data()["feature"].to_list()

    def get_feature_family(features_l: List[str], family: str) -> List[str]:
        return list(filter(lambda el: el.startswith(family), features_l))

    def sum_list_feature(features_l: List[str]) -> Value:
        out: Value = Fe(features_l[0])
        for el in features_l[1:]:
            out = out + Fe(el)
        return out

    def sum_feature_family(features_l: List[str], family: str) -> Value:
        return sum_list_feature(get_feature_family(features_l, family))

    g1 = (
        sum_feature_family(features, "icmp_sum_s_")
        + sum_feature_family(features, "udp_sum_s_")
        + sum_feature_family(features, "tcp_sum_s_")
    ) == (
        sum_feature_family(features, "bytes_in_sum_s_")
        + sum_feature_family(features, "bytes_out_sum_s_")
    )

    g2 = (
        sum_feature_family(features, "icmp_sum_d_")
        + sum_feature_family(features, "udp_sum_d_")
        + sum_feature_family(features, "tcp_sum_d_")
    ) == (
        sum_feature_family(features, "bytes_in_sum_d_")
        + sum_feature_family(features, "bytes_out_sum_d_")
    )

    g_packet_size = []
    for e in ["s", "d"]:
        # -1 cause ignore last OTHER features
        bytes_outs = get_feature_family(features, f"bytes_out_sum_{e}_")[:-1]
        pkts_outs = get_feature_family(features, f"pkts_out_sum_{e}_")[:-1]
        if len(bytes_outs) != len(pkts_outs):
            raise Exception("len(bytes_out) != len(pkts_out)")

        # Tuple of list to list of tuples
        for byte_out, pkts_out in list(zip(bytes_outs, pkts_outs)):
            g = SafeDivision(Fe(byte_out), Fe(pkts_out), Co(0.0)) <= Co(1500)
            g_packet_size.append(g)

    g_min_max_sum = []
    for e_1 in ["bytes_out", "pkts_out", "duration"]:
        for port in [
            "1",
            "3",
            "8",
            "10",
            "21",
            "22",
            "25",
            "53",
            "80",
            "110",
            "123",
            "135",
            "138",
            "161",
            "443",
            "445",
            "993",
            "OTHER",
        ]:
            for e_2 in ["d", "s"]:
                g_min_max_sum.extend(
                    [
                        Fe(f"{e_1}_max_{e_2}_{port}")
                        <= Fe(f"{e_1}_sum_{e_2}_{port}"),
                        Fe(f"{e_1}_min_{e_2}_{port}")
                        <= Fe(f"{e_1}_sum_{e_2}_{port}"),
                        Fe(f"{e_1}_min_{e_2}_{port}")
                        <= Fe(f"{e_1}_max_{e_2}_{port}"),
                    ]
                )

    return [g1, g2] + g_packet_size + g_min_max_sum


def create_dataset() -> Dataset:
    data_source = DownloadFileDataSource(
        url="https://uniluxembourg-my.sharepoint.com/:x:/g/personal/"
        "thibault_simonetto_uni_lu/"
        "Ed5Wox3GpUtChu5ZYzMfACEB3PDRE3fMloUw04MSNiAkTQ?download=1",
        file_data_source=CsvDataSource(
            path="./data/mlc/ctu_13_neris/ctu_13_neris.csv"
        ),
    )
    metadata_source = DownloadFileDataSource(
        url="https://uniluxembourg-my.sharepoint.com/:x:/g/personal/"
        "thibault_simonetto_uni_lu/"
        "ESz5WJPXeF1Fv_i3vVk9cY8B-fqbrYuaaUy_e_iypmZFbQ?download=1",
        file_data_source=CsvDataSource(
            path="./data/mlc/ctu_13_neris/ctu_13_neris_metadata.csv"
        ),
    )
    tasks = [
        Task(
            name="is_botnet",
            task_type="classification",
            evaluation_metric="f1_score",
        )
    ]
    sorter = DefaultIndexSorter()
    splitter = Ctu13Splitter()
    relation_constraints = get_relation_constraints(metadata_source)

    ctu_13_neris = Dataset(
        name="ctu_13_neris",
        data_source=data_source,
        metadata_source=metadata_source,
        tasks=tasks,
        sorter=sorter,
        splitter=splitter,
        relation_constraints=relation_constraints,
    )
    return ctu_13_neris


datasets = [
    {
        "name": "ctu_13_neris",
        "fun_create": create_dataset,
    }
]



from typing import Dict, List

import numpy as np
import numpy.typing as npt
import pandas as pd
from sklearn.model_selection import train_test_split

from mlc.constraints.relation_constraint import (
    BaseRelationConstraint,
    Constant,
    Feature,
    SafeDivision,
)
from mlc.datasets.dataset import (
    CsvDataSource,
    Dataset,
    DateSorter,
    DownloadFileDataSource,
    Splitter,
    Task,
)


class LcldSplitterTime(Splitter):
    def get_splits(self, dataset: Dataset) -> Dict[str, npt.NDArray[np.int_]]:
        _, _, t = dataset.get_x_y_t()
        i_train = np.where(
            (pd.to_datetime("2013-01-01") <= t)
            & (t <= pd.to_datetime("2015-06-30"))
        )[0]
        i_val = np.where(
            (pd.to_datetime("2015-07-01") <= t)
            & (t <= pd.to_datetime("2015-12-31"))
        )[0]
        i_test = np.where(
            (pd.to_datetime("2016-01-01") <= t)
            & (t <= pd.to_datetime("2017-12-31"))
        )[0]
        return {"train": i_train, "val": i_val, "test": i_test}


class LcldSplitterIid(LcldSplitterTime):
    def get_splits(self, dataset: Dataset) -> Dict[str, npt.NDArray[np.int_]]:
        _, y, _ = dataset.get_x_y_t()
        splits = super(LcldSplitterIid, self).get_splits(dataset)
        merged = np.concatenate(
            [splits.get("train"), splits.get("val"), splits.get("test")]
        )
        i_train, i_test = train_test_split(
            merged,
            random_state=100,
            shuffle=True,
            stratify=y[merged],
            test_size=splits["test"].shape[0],
        )
        i_train, i_val = train_test_split(
            i_train,
            random_state=200,
            shuffle=True,
            stratify=y[i_train],
            test_size=splits["val"].shape[0],
        )
        return {"train": i_train, "val": i_val, "test": i_test}


def get_relation_constraints() -> List[BaseRelationConstraint]:

    int_rate = Feature("int_rate") / Constant(1200)
    term = Feature("term")
    installment = Feature("loan_amnt") * (
        (int_rate * ((Constant(1) + int_rate) ** term))
        / ((Constant(1) + int_rate) ** term - Constant(1))
    )
    g1 = Feature("installment") == installment

    g2 = Feature("open_acc") <= Feature("total_acc")

    g3 = Feature("pub_rec_bankruptcies") <= Feature("pub_rec")

    g4 = (Feature("term") == Constant(36)) | (Feature("term") == Constant(60))

    g5 = Feature("ratio_loan_amnt_annual_inc") == (
        Feature("loan_amnt") / Feature("annual_inc")
    )

    g6 = Feature("ratio_open_acc_total_acc") == (
        Feature("open_acc") / Feature("total_acc")
    )

    # g7 was diff_issue_d_earliest_cr_line
    # g7 is not necessary in v2
    # issue_d and d_earliest cr_line are replaced
    # by month_since_earliest_cr_line

    g8 = Feature("ratio_pub_rec_month_since_earliest_cr_line") == (
        Feature("pub_rec") / Feature("month_since_earliest_cr_line")
    )

    g9 = Feature(
        "ratio_pub_rec_bankruptcies_month_since_earliest_cr_line"
    ) == (
        Feature("pub_rec_bankruptcies")
        / Feature("month_since_earliest_cr_line")
    )

    g10 = Feature("ratio_pub_rec_bankruptcies_pub_rec") == SafeDivision(
        Feature("pub_rec_bankruptcies"), Feature("pub_rec"), Constant(-1)
    )

    return [g1, g2, g3, g4, g5, g6, g8, g9, g10]


def is_linked_bool(list_params: List[bool]) -> bool:
    saw_false = False

    for e in list_params:
        saw_false = saw_false or (not e)
        if saw_false and e:
            return False
    return True


def create_dataset(
    split_date: bool, date_1317: bool, simulate_date: bool
) -> Dataset:

    if not is_linked_bool([split_date, date_1317, simulate_date]):
        raise NotImplementedError

    tasks = [
        Task(
            name="charged_off",
            task_type="classification",
            evaluation_metric="f1_score",
        )
    ]
    sorter = DateSorter(date_col="issue_d")
    data_source = DownloadFileDataSource(
        url="https://uniluxembourg-my.sharepoint.com/:x:/g/personal/"
        "thibault_simonetto_uni_lu/"
        "EeKpMXnQNo9CuRaLcHNjiX0B4Tf2H_HV3OKmlqvwbZZ-aA?download=1",
        file_data_source=CsvDataSource(path="./data/mlc/lcld_v2/lcld_v2.csv"),
    )
    metadata_source = DownloadFileDataSource(
        url="https://uniluxembourg-my.sharepoint.com/:x:/g/personal/"
        "thibault_simonetto_uni_lu/"
        "EfIRH-UBrW1BiYwvSjlyZIIBFbOiALcxUzdZ5T11qhILlw?download=1",
        file_data_source=CsvDataSource(
            path="./data/mlc/lcld_v2/lcld_v2_metadata.csv"
        ),
    )

    if split_date:
        name = "lcld_time"
        splitter = LcldSplitterTime()
    else:
        name = "lcld_iid"
        splitter = LcldSplitterIid()

    if date_1317:
        name = "lcld_201317_time"

        data_source = DownloadFileDataSource(
            url="https://uniluxembourg-my.sharepoint.com/:x:/g/personal/"
            "thibault_simonetto_uni_lu/"
            "EYRPThLbct9DlYcWsKUbRlgBES9QO2Nk8hZs1cmrQ7tzMQ?download=1",
            file_data_source=CsvDataSource(path="./data/mlc/lcld_201317/lcld_201317.csv"),
        )
        metadata_source = DownloadFileDataSource(
            url="https://uniluxembourg-my.sharepoint.com/:x:/g/personal/"
            "thibault_simonetto_uni_lu/"
            "EUS4y7I7-FBOrMKV7tb7tmsBUCHdAkCBKXaqEu4V-YuZeQ?download=1",
            file_data_source=CsvDataSource(
                path="./data/mlc/lcld_201317/lcld_201317_metadata.csv"
            ),
        )

        if simulate_date:
            name = "lcld_201317_ds_time"

            data_source = DownloadFileDataSource(
                url="https://uniluxembourg-my.sharepoint.com/:x:/g/personal/"
                "thibault_simonetto_uni_lu/"
                "EVMe8n0hiEJCuf3Y1oO7BmwBBgL0uzAAizboz-aSlcxwBA?download=1",
                file_data_source=CsvDataSource(path="./data/mlc/lcld_201317_ds/lcld_201317_ds.csv"),
            )
            metadata_source = DownloadFileDataSource(
                url="https://uniluxembourg-my.sharepoint.com/:x:/g/personal/"
                "thibault_simonetto_uni_lu/"
                "EZ9ttVvRzSZPjGtcnmxUwYEBZsxhByGttYAU9GGzX5kosQ?download=1",
                file_data_source=CsvDataSource(
                    path="./data/mlc/lcld_201317_ds/lcld_201317_ds_metadata.csv"
                ),
            )

    dataset = Dataset(
        name=name,
        data_source=data_source,
        metadata_source=metadata_source,
        tasks=tasks,
        sorter=sorter,
        splitter=splitter,
        relation_constraints=get_relation_constraints(),
    )
    return dataset


# V2 for backward compatibility
datasets = [
    {
        "name": "lcld_iid",
        "fun_create": lambda: create_dataset(False, False, False),
    },
    {
        "name": "lcld_v2_iid",
        "fun_create": lambda: create_dataset(False, False, False),
    },
    {
        "name": "lcld_time",
        "fun_create": lambda: create_dataset(True, False, False),
    },
    {
        "name": "lcld_v2_time",
        "fun_create": lambda: create_dataset(True, False, False),
    },
    {
        "name": "lcld_201317_time",
        "fun_create": lambda: create_dataset(True, True, False),
    },
    {
        "name": "lcld_201317_ds_time",
        "fun_create": lambda: create_dataset(True, True, True),
    },
]



from mlc.logging import XP
import json
import pathlib
import time
from types import SimpleNamespace
from typing import Any, Dict, List, Optional, Union

import numpy as np
import numpy.typing as npt
import torch
from sklearn.model_selection import KFold, StratifiedKFold

from mlc.metrics.compute import compute_metric
from mlc.metrics.metric_factory import create_metric
from mlc.typing import NDNumber


def cut_in_batch(
    arr: npt.NDArray[Any],
    n_desired_batch: int = 1,
    batch_size: Optional[int] = None,
) -> List[npt.NDArray[Any]]:

    if batch_size is None:
        n_batch = min(n_desired_batch, len(arr))
    else:
        n_batch = np.ceil(len(arr) / batch_size)
    batches_i = np.array_split(np.arange(arr.shape[0]), n_batch)

    return [arr[batch_i] for batch_i in batches_i]


def dict2obj(d: Dict[str, Any]) -> Any:
    dumped_data = json.dumps(d)
    result = json.loads(
        dumped_data, object_hook=lambda x: SimpleNamespace(**x)
    )
    return result


def parent_exists(path: str) -> str:
    pathlib.Path(path).parent.mkdir(parents=True, exist_ok=True)
    return path


class Timer:
    def __init__(self):
        self.save_times = []
        self.start_time = 0

    def start(self):
        self.start_time = time.process_time()

    def end(self):
        end_time = time.process_time()
        self.save_times.append(end_time - self.start_time)

    def get_average_time(self):
        return np.mean(self.save_times)


def cross_validation(model, X, y, metric, args):
    # Record some statistics and metrics
    sc = create_metric(metric)
    train_timer = Timer()
    test_timer = Timer()

    if args.objective == "regression":
        kf = KFold(
            n_splits=args.num_splits,
            shuffle=args.shuffle,
            random_state=args.seed,
        )
    elif args.objective == "classification" or args.objective == "binary":
        kf = StratifiedKFold(
            n_splits=args.num_splits,
            shuffle=args.shuffle,
            random_state=args.seed,
        )
    else:
        raise NotImplementedError(
            "Objective" + args.objective + "is not yet implemented."
        )

    for i, (train_index, test_index) in enumerate(kf.split(X, y)):

        experiment = XP(vars(args))

        X_train, X_test = X[train_index], X[test_index]
        y_train, y_test = y[train_index], y[test_index]

        if args.objective == "binary":
            y_train = np.column_stack([1 - y_train, y_train])
            y_test = np.column_stack([1 - y_test, y_test])

        # Create a new unfitted version of the model
        curr_model = model.clone()
        curr_model.set_experiment(experiment)

        # Train model
        train_timer.start()
        curr_model.fit(
            X_train, y_train, X_test, y_test
        )  # X_val, y_val)
        train_timer.end()

        # Test model
        test_timer.start()
        metric_val = compute_metric(curr_model, sc, X_test, y_test)
        test_timer.end()

        print(metric_val)
        experiment.log_metric(sc.metric_name,metric_val)
        experiment.end()


    return metric_val, (
        train_timer.get_average_time(),
        test_timer.get_average_time(),
    )


def round_func_BPDA(input):
    # This is equivalent to replacing round function (non-differentiable) with
    # an identity function (differentiable) only when backward.
    forward_value = torch.round(input)
    out = input.clone()
    out.data = forward_value.data
    return out


def to_torch_number(tensor: Union[torch.Tensor, NDNumber]) -> torch.Tensor:
    if isinstance(tensor, torch.Tensor):
        return tensor

    elif isinstance(tensor, np.ndarray):
        return torch.from_numpy(tensor)

    raise NotImplementedError(f"Unsupported type: {type(tensor)}")


def to_numpy_number(tensor: Union[torch.Tensor, NDNumber]) -> NDNumber:
    if isinstance(tensor, torch.Tensor):
        return tensor.detach().cpu().numpy()

    elif isinstance(tensor, np.ndarray):
        return tensor

    raise NotImplementedError(f"Unsupported type: {type(tensor)}")



from typing import Union

import numpy as np
import numpy.typing as npt

NDNumber = npt.NDArray[Union[np.float_, np.int_]]
NDFloat = npt.NDArray[np.float_]
NDInt = npt.NDArray[np.int_]
NDBool = npt.NDArray[np.bool_]



import torch
import torch.nn.functional as F
from torch import nn, einsum
import numpy as np
from einops import rearrange


# helpers

def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


def ff_encodings(x, B):
    x_proj = (2. * np.pi * x.unsqueeze(-1)) @ B.t()
    return torch.cat([torch.sin(x_proj), torch.cos(x_proj)], dim=-1)


# classes

class Residual(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x, **kwargs):
        return self.fn(x, **kwargs) + x


class PreNorm(nn.Module):
    def __init__(self, dim, fn):
        super().__init__()
        self.norm = nn.LayerNorm(dim)
        self.fn = fn

    def forward(self, x, **kwargs):
        return self.fn(self.norm(x), **kwargs)


# attention

class GEGLU(nn.Module):
    def forward(self, x):
        x, gates = x.chunk(2, dim=-1)
        return x * F.gelu(gates)


class FeedForward(nn.Module):
    def __init__(self, dim, mult=4, dropout=0.):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(dim, dim * mult * 2),
            GEGLU(),
            nn.Dropout(dropout),
            nn.Linear(dim * mult, dim)
        )

    def forward(self, x, **kwargs):
        return self.net(x)


class Attention(nn.Module):
    def __init__(
            self,
            dim,
            heads=8,
            dim_head=16,
            dropout=0.
    ):
        super().__init__()
        inner_dim = dim_head * heads
        self.heads = heads
        self.scale = dim_head ** -0.5

        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias=False)
        self.to_out = nn.Linear(inner_dim, dim)

        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        h = self.heads
        q, k, v = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h=h), (q, k, v))
        sim = einsum('b h i d, b h j d -> b h i j', q, k) * self.scale
        attn = sim.softmax(dim=-1)
        out = einsum('b h i j, b h j d -> b h i d', attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)', h=h)
        return self.to_out(out)


class RowColTransformer(nn.Module):
    def __init__(self, num_tokens, dim, nfeats, depth, heads, dim_head, attn_dropout, ff_dropout, style='col'):
        super().__init__()
        self.embeds = nn.Embedding(num_tokens, dim)
        self.layers = nn.ModuleList([])
        self.mask_embed = nn.Embedding(nfeats, dim)
        self.style = style
        for _ in range(depth):
            if self.style == 'colrow':
                self.layers.append(nn.ModuleList([
                    PreNorm(dim, Residual(Attention(dim, heads=heads, dim_head=dim_head, dropout=attn_dropout))),
                    PreNorm(dim, Residual(FeedForward(dim, dropout=ff_dropout))),
                    PreNorm(dim * nfeats,
                            Residual(Attention(dim * nfeats, heads=heads, dim_head=64, dropout=attn_dropout))),
                    PreNorm(dim * nfeats, Residual(FeedForward(dim * nfeats, dropout=ff_dropout))),
                ]))
            else:
                self.layers.append(nn.ModuleList([
                    PreNorm(dim * nfeats,
                            Residual(Attention(dim * nfeats, heads=heads, dim_head=64, dropout=attn_dropout))),
                    PreNorm(dim * nfeats, Residual(FeedForward(dim * nfeats, dropout=ff_dropout))),
                ]))

    def forward(self, x, x_cont=None, mask=None):
        if x_cont is not None:
            x = torch.cat((x, x_cont), dim=1)
        _, n, _ = x.shape
        if self.style == 'colrow':
            for attn1, ff1, attn2, ff2 in self.layers:
                x = attn1(x)
                x = ff1(x)
                x = rearrange(x, 'b n d -> 1 b (n d)')
                x = attn2(x)
                x = ff2(x)
                x = rearrange(x, '1 b (n d) -> b n d', n=n)
        else:
            for attn1, ff1 in self.layers:
                x = rearrange(x, 'b n d -> 1 b (n d)')
                x = attn1(x)
                x = ff1(x)
                x = rearrange(x, '1 b (n d) -> b n d', n=n)
        return x


# transformer
class Transformer(nn.Module):
    def __init__(self, num_tokens, dim, depth, heads, dim_head, attn_dropout, ff_dropout):
        super().__init__()
        self.layers = nn.ModuleList([])

        for _ in range(depth):
            self.layers.append(nn.ModuleList([
                PreNorm(dim, Residual(Attention(dim, heads=heads, dim_head=dim_head, dropout=attn_dropout))),
                PreNorm(dim, Residual(FeedForward(dim, dropout=ff_dropout))),
            ]))

    def forward(self, x, x_cont=None):
        if x_cont is not None:
            x = torch.cat((x, x_cont), dim=1)
        for attn, ff in self.layers:
            x = attn(x)
            x = ff(x)
        return x


# mlp
class MLP(nn.Module):
    def __init__(self, dims, act=None):
        super().__init__()
        dims_pairs = list(zip(dims[:-1], dims[1:]))
        layers = []
        for ind, (dim_in, dim_out) in enumerate(dims_pairs):
            is_last = ind >= (len(dims) - 1)
            linear = nn.Linear(dim_in, dim_out)
            layers.append(linear)

            if is_last:
                continue
            if act is not None:
                layers.append(act)

        self.mlp = nn.Sequential(*layers)

    def forward(self, x):
        return self.mlp(x)


class simple_MLP(nn.Module):
    def __init__(self, dims):
        super(simple_MLP, self).__init__()
        self.layers = nn.Sequential(
            nn.Linear(dims[0], dims[1]),
            nn.ReLU(),
            nn.Linear(dims[1], dims[2])
        )

    def forward(self, x):
        if len(x.shape) == 1:
            x = x.view(x.size(0), -1)
        x = self.layers(x)
        return x


# main class

class TabAttention(nn.Module):
    def __init__(
            self,
            *,
            categories,
            num_continuous,
            dim,
            depth,
            heads,
            dim_head=16,
            dim_out=1,
            mlp_hidden_mults=(4, 2),
            mlp_act=None,
            num_special_tokens=1,
            continuous_mean_std=None,
            attn_dropout=0.,
            ff_dropout=0.,
            lastmlp_dropout=0.,
            cont_embeddings='MLP',
            scalingfactor=10,
            attentiontype='col'
    ):
        super().__init__()
        assert all(map(lambda n: n > 0, categories)), 'number of each category must be positive'

        # categories related calculations
        self.num_categories = len(categories)
        self.num_unique_categories = sum(categories)

        # create category embeddings table

        self.num_special_tokens = num_special_tokens
        self.total_tokens = self.num_unique_categories + num_special_tokens

        # for automatically offsetting unique category ids to the correct position in the categories embedding table
        categories_offset = F.pad(torch.tensor(list(categories)), (1, 0), value=num_special_tokens)
        categories_offset = categories_offset.cumsum(dim=-1)[:-1]

        self.register_buffer('categories_offset', categories_offset)

        self.norm = nn.LayerNorm(num_continuous)
        self.num_continuous = num_continuous
        self.dim = dim
        self.cont_embeddings = cont_embeddings
        self.attentiontype = attentiontype

        if self.cont_embeddings == 'MLP':
            self.simple_MLP = nn.ModuleList([simple_MLP([1, 100, self.dim]) for _ in range(self.num_continuous)])
            input_size = (dim * self.num_categories) + (dim * num_continuous)
            nfeats = self.num_categories + num_continuous
        else:
            print('Continous features are not passed through attention')
            input_size = (dim * self.num_categories) + num_continuous
            nfeats = self.num_categories

            # transformer
        if attentiontype == 'col':
            self.transformer = Transformer(
                num_tokens=self.total_tokens,
                dim=dim,
                depth=depth,
                heads=heads,
                dim_head=dim_head,
                attn_dropout=attn_dropout,
                ff_dropout=ff_dropout
            )
        elif attentiontype in ['row', 'colrow']:
            self.transformer = RowColTransformer(
                num_tokens=self.total_tokens,
                dim=dim,
                nfeats=nfeats,
                depth=depth,
                heads=heads,
                dim_head=dim_head,
                attn_dropout=attn_dropout,
                ff_dropout=ff_dropout,
                style=attentiontype
            )

        l = input_size // 8
        hidden_dimensions = list(map(lambda t: l * t, mlp_hidden_mults))
        all_dimensions = [input_size, *hidden_dimensions, dim_out]

        self.mlp = MLP(all_dimensions, act=mlp_act)
        self.embeds = nn.Embedding(self.total_tokens, self.dim)  # .to(device)

        cat_mask_offset = F.pad(torch.Tensor(self.num_categories).fill_(2).type(torch.int8), (1, 0), value=0)
        cat_mask_offset = cat_mask_offset.cumsum(dim=-1)[:-1]

        con_mask_offset = F.pad(torch.Tensor(self.num_continuous).fill_(2).type(torch.int8), (1, 0), value=0)
        con_mask_offset = con_mask_offset.cumsum(dim=-1)[:-1]

        self.register_buffer('cat_mask_offset', cat_mask_offset)
        self.register_buffer('con_mask_offset', con_mask_offset)

        self.mask_embeds_cat = nn.Embedding(self.num_categories * 2, self.dim)
        self.mask_embeds_cont = nn.Embedding(self.num_continuous * 2, self.dim)

    def forward(self, x_categ, x_cont, x_categ_enc, x_cont_enc):
        device = x_categ.device
        if self.attentiontype == 'justmlp':
            if x_categ.shape[-1] > 0:
                flat_categ = x_categ.flatten(1).to(device)
                x = torch.cat((flat_categ, x_cont.flatten(1).to(device)), dim=-1)
            else:
                x = x_cont.clone()
        else:
            if self.cont_embeddings == 'MLP':
                x = self.transformer(x_categ_enc, x_cont_enc.to(device))
            else:
                if x_categ.shape[-1] <= 0:
                    x = x_cont.clone()
                else:
                    flat_categ = self.transformer(x_categ_enc).flatten(1)
                    x = torch.cat((flat_categ, x_cont), dim=-1)
        flat_x = x.flatten(1)
        return self.mlp(flat_x)



"""
Code was adapted from https://github.com/Yura52/rtdl
"""

import math
from typing import Callable, List, Type, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim
from torch import Tensor

ModuleType = Union[str, Callable[..., nn.Module]]


class SiLU(nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(x)


def timestep_embedding(timesteps, dim, max_period=10000):
    """
    Create sinusoidal timestep embeddings.

    :param timesteps: a 1-D Tensor of N indices, one per batch element.
                      These may be fractional.
    :param dim: the dimension of the output.
    :param max_period: controls the minimum frequency of the embeddings.
    :return: an [N x dim] Tensor of positional embeddings.
    """
    half = dim // 2
    freqs = torch.exp(
        -math.log(max_period) * torch.arange(start=0, end=half, dtype=torch.float32) / half
    ).to(device=timesteps.device)
    args = timesteps[:, None].float() * freqs[None]
    embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1)
    if dim % 2:
        embedding = torch.cat([embedding, torch.zeros_like(embedding[:, :1])], dim=-1)
    return embedding


def _is_glu_activation(activation: ModuleType):
    return (
            isinstance(activation, str)
            and activation.endswith('GLU')
            or activation in [ReGLU, GEGLU]
    )


def _all_or_none(values):
    assert all(x is None for x in values) or all(x is not None for x in values)


def reglu(x: Tensor) -> Tensor:
    """The ReGLU activation function from [1].
    References:
        [1] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """
    assert x.shape[-1] % 2 == 0
    a, b = x.chunk(2, dim=-1)
    return a * F.relu(b)


def geglu(x: Tensor) -> Tensor:
    """The GEGLU activation function from [1].
    References:
        [1] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """
    assert x.shape[-1] % 2 == 0
    a, b = x.chunk(2, dim=-1)
    return a * F.gelu(b)


class ReGLU(nn.Module):
    """The ReGLU activation function from [shazeer2020glu].

    Examples:
        .. testcode::

            module = ReGLU()
            x = torch.randn(3, 4)
            assert module(x).shape == (3, 2)

    References:
        * [shazeer2020glu] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """

    def forward(self, x: Tensor) -> Tensor:
        return reglu(x)


class GEGLU(nn.Module):
    """The GEGLU activation function from [shazeer2020glu].

    Examples:
        .. testcode::

            module = GEGLU()
            x = torch.randn(3, 4)
            assert module(x).shape == (3, 2)

    References:
        * [shazeer2020glu] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """

    def forward(self, x: Tensor) -> Tensor:
        return geglu(x)


def _make_nn_module(module_type: ModuleType, *args) -> nn.Module:
    return (
        (
            ReGLU()
            if module_type == 'ReGLU'
            else GEGLU()
            if module_type == 'GEGLU'
            else getattr(nn, module_type)(*args)
        )
        if isinstance(module_type, str)
        else module_type(*args)
    )


class MLP(nn.Module):
    """The MLP model used in [gorishniy2021revisiting].

    The following scheme describes the architecture:

    .. code-block:: text

          MLP: (in) -> Block -> ... -> Block -> Linear -> (out)
        Block: (in) -> Linear -> Activation -> Dropout -> (out)

    Examples:
        .. testcode::

            x = torch.randn(4, 2)
            module = MLP.make_baseline(x.shape[1], [3, 5], 0.1, 1)
            assert module(x).shape == (len(x), 1)

    References:
        * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
    """

    class Block(nn.Module):
        """The main building block of `MLP`."""

        def __init__(
                self,
                *,
                d_in: int,
                d_out: int,
                bias: bool,
                activation: ModuleType,
                dropout: float,
        ) -> None:
            super().__init__()
            self.linear = nn.Linear(d_in, d_out, bias)
            self.activation = _make_nn_module(activation)
            self.dropout = nn.Dropout(dropout)

        def forward(self, x: Tensor) -> Tensor:
            return self.dropout(self.activation(self.linear(x)))

    def __init__(
            self,
            *,
            d_in: int,
            d_layers: List[int],
            dropouts: Union[float, List[float]],
            activation: Union[str, Callable[[], nn.Module]],
            d_out: int,
    ) -> None:
        """
        Note:
            `make_baseline` is the recommended constructor.
        """
        super().__init__()
        if isinstance(dropouts, float):
            dropouts = [dropouts] * len(d_layers)
        assert len(d_layers) == len(dropouts)
        assert activation not in ['ReGLU', 'GEGLU']

        self.blocks = nn.ModuleList(
            [
                MLP.Block(
                    d_in=d_layers[i - 1] if i else d_in,
                    d_out=d,
                    bias=True,
                    activation=activation,
                    dropout=dropout,
                )
                for i, (d, dropout) in enumerate(zip(d_layers, dropouts))
            ]
        )
        self.head = nn.Linear(d_layers[-1] if d_layers else d_in, d_out)

    @classmethod
    def make_baseline(
            cls: Type['MLP'],
            d_in: int,
            d_layers: List[int],
            dropout: float,
            d_out: int,
    ) -> 'MLP':
        """Create a "baseline" `MLP`.

        This variation of MLP was used in [gorishniy2021revisiting]. Features:

        * :code:`Activation` = :code:`ReLU`
        * all linear layers except for the first one and the last one are of the same dimension
        * the dropout rate is the same for all dropout layers

        Args:
            d_in: the input size
            d_layers: the dimensions of the linear layers. If there are more than two
                layers, then all of them except for the first and the last ones must
                have the same dimension. Valid examples: :code:`[]`, :code:`[8]`,
                :code:`[8, 16]`, :code:`[2, 2, 2, 2]`, :code:`[1, 2, 2, 4]`. Invalid
                example: :code:`[1, 2, 3, 4]`.
            dropout: the dropout rate for all hidden layers
            d_out: the output size
        Returns:
            MLP

        References:
            * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
        """
        assert isinstance(dropout, float)
        if len(d_layers) > 2:
            assert len(set(d_layers[1:-1])) == 1, (
                'if d_layers contains more than two elements, then'
                ' all elements except for the first and the last ones must be equal.'
            )
        return MLP(
            d_in=d_in,
            d_layers=d_layers,  # type: ignore
            dropouts=dropout,
            activation='ReLU',
            d_out=d_out,
        )

    def forward(self, x: Tensor) -> Tensor:
        x = x.float()
        for block in self.blocks:
            x = block(x)
        x = self.head(x)
        return x


class ResNet(nn.Module):
    """The ResNet model used in [gorishniy2021revisiting].
    The following scheme describes the architecture:
    .. code-block:: text
        ResNet: (in) -> Linear -> Block -> ... -> Block -> Head -> (out)
                 |-> Norm -> Linear -> Activation -> Dropout -> Linear -> Dropout ->|
                 |                                                                  |
         Block: (in) ------------------------------------------------------------> Add -> (out)
          Head: (in) -> Norm -> Activation -> Linear -> (out)
    Examples:
        .. testcode::
            x = torch.randn(4, 2)
            module = ResNet.make_baseline(
                d_in=x.shape[1],
                n_blocks=2,
                d_main=3,
                d_hidden=4,
                dropout_first=0.25,
                dropout_second=0.0,
                d_out=1
            )
            assert module(x).shape == (len(x), 1)
    References:
        * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
    """

    class Block(nn.Module):
        """The main building block of `ResNet`."""

        def __init__(
                self,
                *,
                d_main: int,
                d_hidden: int,
                bias_first: bool,
                bias_second: bool,
                dropout_first: float,
                dropout_second: float,
                normalization: ModuleType,
                activation: ModuleType,
                skip_connection: bool,
        ) -> None:
            super().__init__()
            self.normalization = _make_nn_module(normalization, d_main)
            self.linear_first = nn.Linear(d_main, d_hidden, bias_first)
            self.activation = _make_nn_module(activation)
            self.dropout_first = nn.Dropout(dropout_first)
            self.linear_second = nn.Linear(d_hidden, d_main, bias_second)
            self.dropout_second = nn.Dropout(dropout_second)
            self.skip_connection = skip_connection

        def forward(self, x: Tensor) -> Tensor:
            x_input = x
            x = self.normalization(x)
            x = self.linear_first(x)
            x = self.activation(x)
            x = self.dropout_first(x)
            x = self.linear_second(x)
            x = self.dropout_second(x)
            if self.skip_connection:
                x = x_input + x
            return x

    class Head(nn.Module):
        """The final module of `ResNet`."""

        def __init__(
                self,
                *,
                d_in: int,
                d_out: int,
                bias: bool,
                normalization: ModuleType,
                activation: ModuleType,
        ) -> None:
            super().__init__()
            self.normalization = _make_nn_module(normalization, d_in)
            self.activation = _make_nn_module(activation)
            self.linear = nn.Linear(d_in, d_out, bias)

        def forward(self, x: Tensor) -> Tensor:
            if self.normalization is not None:
                x = self.normalization(x)
            x = self.activation(x)
            x = self.linear(x)
            return x

    def __init__(
            self,
            *,
            d_in: int,
            n_blocks: int,
            d_main: int,
            d_hidden: int,
            dropout_first: float,
            dropout_second: float,
            normalization: ModuleType,
            activation: ModuleType,
            d_out: int,
    ) -> None:
        """
        Note:
            `make_baseline` is the recommended constructor.
        """
        super().__init__()

        self.first_layer = nn.Linear(d_in, d_main)
        if d_main is None:
            d_main = d_in
        self.blocks = nn.Sequential(
            *[
                ResNet.Block(
                    d_main=d_main,
                    d_hidden=d_hidden,
                    bias_first=True,
                    bias_second=True,
                    dropout_first=dropout_first,
                    dropout_second=dropout_second,
                    normalization=normalization,
                    activation=activation,
                    skip_connection=True,
                )
                for _ in range(n_blocks)
            ]
        )
        self.head = ResNet.Head(
            d_in=d_main,
            d_out=d_out,
            bias=True,
            normalization=normalization,
            activation=activation,
        )

    @classmethod
    def make_baseline(
            cls: Type['ResNet'],
            *,
            d_in: int,
            n_blocks: int,
            d_main: int,
            d_hidden: int,
            dropout_first: float,
            dropout_second: float,
            d_out: int,
    ) -> 'ResNet':
        """Create a "baseline" `ResNet`.
        This variation of ResNet was used in [gorishniy2021revisiting]. Features:
        * :code:`Activation` = :code:`ReLU`
        * :code:`Norm` = :code:`BatchNorm1d`
        Args:
            d_in: the input size
            n_blocks: the number of Blocks
            d_main: the input size (or, equivalently, the output size) of each Block
            d_hidden: the output size of the first linear layer in each Block
            dropout_first: the dropout rate of the first dropout layer in each Block.
            dropout_second: the dropout rate of the second dropout layer in each Block.
        References:
            * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
        """
        return cls(
            d_in=d_in,
            n_blocks=n_blocks,
            d_main=d_main,
            d_hidden=d_hidden,
            dropout_first=dropout_first,
            dropout_second=dropout_second,
            normalization='BatchNorm1d',
            activation='ReLU',
            d_out=d_out,
        )

    def forward(self, x: Tensor) -> Tensor:
        x = x.float()
        x = self.first_layer(x)
        x = self.blocks(x)
        x = self.head(x)
        return x


#### For diffusion

class MLPDiffusion(nn.Module):
    def __init__(self, d_in, num_classes, is_y_cond, rtdl_params, dim_t=128):
        super().__init__()
        self.dim_t = dim_t
        self.num_classes = num_classes
        self.is_y_cond = is_y_cond

        # d0 = rtdl_params['d_layers'][0]

        rtdl_params['d_in'] = dim_t
        rtdl_params['d_out'] = d_in

        self.mlp = MLP.make_baseline(**rtdl_params)

        if self.num_classes > 0 and is_y_cond:
            self.label_emb = nn.Embedding(self.num_classes, dim_t)
        elif self.num_classes == 0 and is_y_cond:
            self.label_emb = nn.Linear(1, dim_t)

        self.proj = nn.Linear(d_in, dim_t)
        self.time_embed = nn.Sequential(
            nn.Linear(dim_t, dim_t),
            nn.SiLU(),
            nn.Linear(dim_t, dim_t)
        )

    def forward(self, x, timesteps, y=None):
        emb = self.time_embed(timestep_embedding(timesteps, self.dim_t))
        if self.is_y_cond and y is not None:
            if self.num_classes > 0:
                y = y.squeeze()
            else:
                y = y.resize(y.size(0), 1).float()
            emb += F.silu(self.label_emb(y))
        x = self.proj(x) + emb
        return self.mlp(x)


class ResNetDiffusion(nn.Module):
    def __init__(self, d_in, num_classes, rtdl_params, dim_t=256):
        super().__init__()
        self.dim_t = dim_t
        self.num_classes = num_classes

        rtdl_params['d_in'] = d_in
        rtdl_params['d_out'] = d_in
        rtdl_params['emb_d'] = dim_t
        self.resnet = ResNet.make_baseline(**rtdl_params)

        if self.num_classes > 0:
            self.label_emb = nn.Embedding(self.num_classes, dim_t)

        self.time_embed = nn.Sequential(
            nn.Linear(dim_t, dim_t),
            nn.SiLU(),
            nn.Linear(dim_t, dim_t)
        )

    def forward(self, x, timesteps, y=None):
        emb = self.time_embed(timestep_embedding(timesteps, self.dim_t))
        if y is not None and self.num_classes > 0:
            emb += self.label_emb(y.squeeze())
        return self.resnet(x, emb)



from typing import List, Optional, Union

import numpy as np
import torch

from mlc.constraints.backend import Backend


class PytorchBackend(Backend):
    def __init__(self, eps: float = 0.000001, device="cpu") -> None:
        self.eps = torch.tensor(eps)
        self.device = device

    def set_device(self, device: str) -> None:
        self.device = device

    def get_device(self) -> str:
        return self.device

    def get_eps(self) -> torch.Tensor:
        return self.eps

    def get_zeros(self, operands: List[torch.Tensor]) -> torch.Tensor:
        i = np.argmax([op.shape[-1] for op in operands])
        return torch.zeros(operands[i].shape, dtype=operands[i].dtype).to(
            operands[i].device
        )

    # Values
    def constant(self, value: Union[int, float]) -> torch.Tensor:
        return torch.tensor(np.array([value])).to(self.device)

    def feature(self, x: torch.Tensor, feature_id: int) -> torch.Tensor:
        return x[:, feature_id]

    # Math operations

    def math_operation(
        self,
        operator: str,
        left_operand: torch.Tensor,
        right_operand: torch.Tensor,
    ) -> torch.Tensor:
        if operator == "+":
            return left_operand + right_operand
        elif operator == "-":
            return left_operand - right_operand
        elif operator == "*":
            return left_operand * right_operand
        elif operator == "/":
            return left_operand / right_operand
        elif operator == "**":
            return left_operand**right_operand
        elif operator == "%":
            return left_operand % right_operand
        else:
            raise ValueError(f"Unknown operator: {operator}")

    def safe_division(
        self,
        dividend: torch.Tensor,
        divisor: torch.Tensor,
        safe_value: torch.Tensor,
    ) -> torch.Tensor:
        # This new version does not break gradient computation

        out = torch.full_like(dividend, fill_value=float(safe_value))
        mask = torch.abs(divisor) > self.eps
        out[mask] = dividend[mask] / divisor[mask]
        if torch.isinf(out).float().sum() > 0:
            for _ in range(12):
                print("------------------------ INF ------------------------")
        return out

    def log(
        self, operand: torch.Tensor, safe_value: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        if safe_value is not None:
            return torch.where(operand > 0, torch.log(operand), safe_value)
        return torch.log(operand)

    def many_sum(self, operands: List[torch.Tensor]) -> torch.Tensor:
        zeros = self.get_zeros(operands)
        parsed_operands = []
        for op in operands:
            if op.shape[-1] == 1:
                parsed_operands.append(zeros + op)
            else:
                parsed_operands.append(op.to(operands[0].device))

        return torch.sum(torch.stack(parsed_operands), dim=0)

    def less_equal_constraint(
        self, left_operand: torch.Tensor, right_operand: torch.Tensor
    ) -> torch.Tensor:

        zeros = self.get_zeros([left_operand, right_operand]).to(
            left_operand.device
        )
        substraction = left_operand - right_operand
        bound_zero = torch.max(torch.stack([zeros, substraction]), dim=0)[0]

        return bound_zero

    def less_constraint(
        self, left_operand: torch.Tensor, right_operand: torch.Tensor
    ) -> torch.Tensor:
        zeros = self.get_zeros([left_operand, right_operand])
        substraction = (left_operand + self.eps) - right_operand
        bound_zero = torch.max(torch.stack([zeros, substraction]), dim=0)[0]
        return bound_zero

    def equal_constraint(
        self, left_operand: torch.Tensor, right_operand: torch.Tensor
    ) -> torch.Tensor:
        return torch.abs(left_operand - right_operand)

    def or_constraint(self, operands: List[torch.Tensor]) -> torch.Tensor:
        return torch.min(torch.stack(operands), dim=0).values

    def and_constraint(self, operands: List[torch.Tensor]) -> torch.Tensor:
        return torch.sum(torch.stack(operands), dim=0)

    def count(
        self, operands: List[torch.Tensor], inverse: bool
    ) -> torch.Tensor:
        if inverse:
            return torch.sum(
                torch.stack([(op != 0).float() for op in operands]),
                dim=0,
            )
        else:
            return torch.sum(
                torch.stack([(op == 0).float() for op in operands]),
                dim=0,
            )



from mlc.dataloaders.dataloader import BaseDataLoader


class SDVDataLoader(BaseDataLoader):
    def __init__(self, num_workers: int = 4, pretrained_path: str = None, save_path: str = None) -> None:
        self.num_workers = num_workers
        if pretrained_path is None:
            print("fitting generator")
            self.generator.fit(self.data)
        else:
            self.load(pretrained_path)
        if save_path is not None:
            self.save(save_path)

    def load(self, pretrained_path: str):
        self.generator._model.load(pretrained_path)

    def save(self, save_path: str):
        self.generator._model.save(save_path)



import logging
from dataclasses import dataclass
from typing import Any, List, Optional, Tuple

import numpy as np
import numpy.typing as npt
import pandas as pd

from mlc.constraints.relation_constraint import BaseRelationConstraint
from mlc.typing import NDFloat


@dataclass
class Constraints:
    feature_types: npt.NDArray[Any]
    mutable_features: npt.NDArray[Any]
    lower_bounds: npt.NDArray[Any]
    upper_bounds: npt.NDArray[Any]
    relation_constraints: List[BaseRelationConstraint]
    feature_names: Optional[List[str]] = None


def get_constraints_from_file(
    features_path: str,
    relation_constraints: List[BaseRelationConstraint],
    col_filter: Optional[List[str]] = None,
) -> Constraints:
    features = pd.read_csv(features_path)
    return get_constraints_from_metadata(
        features, relation_constraints, col_filter
    )


def get_constraints_from_metadata(
    metadata: pd.DataFrame,
    relation_constraints: List[BaseRelationConstraint],
    col_filter: Optional[List[str]] = None,
) -> Constraints:
    features = metadata
    if col_filter is not None:
        features = features[features["feature"].isin(col_filter)]
    feature_type = features["type"].to_numpy()
    mutable_mask = features["mutable"].to_numpy()
    feature_min = features["min"].to_numpy()
    feature_max = features["max"].to_numpy()
    feature_names = features["feature"].to_numpy()
    return Constraints(
        feature_type,
        mutable_mask,
        feature_min,
        feature_max,
        relation_constraints,
        feature_names,
    )


def get_feature_min_max(
    constraints: Constraints, x: npt.NDArray[Any]
) -> Tuple[NDFloat, NDFloat]:

    # Todo: Implement a faster method than this recursive solution
    if (x is not None) and (x.ndim == 2):
        out = [get_feature_min_max(constraints, x0) for x0 in x]
        feature_min = np.array([e[0] for e in out])
        feature_max = np.array([e[1] for e in out])
        return feature_min, feature_max

    lower_bounds, upper_bounds = (
        constraints.lower_bounds,
        constraints.upper_bounds,
    )
    # By default min and max are the extreme values
    feature_min = np.full(lower_bounds.shape, np.nan)
    feature_max = np.full(upper_bounds.shape, np.nan)

    # Creating the mask of value that should be provided by input
    min_dynamic = lower_bounds.astype(str) == "dynamic"
    max_dynamic = upper_bounds.astype(str) == "dynamic"

    # Replace de non-dynamic value by the value provided in the definition
    feature_min[~min_dynamic] = lower_bounds[~min_dynamic]
    feature_max[~max_dynamic] = upper_bounds[~max_dynamic]

    # If the dynamic input was provided, replace value for output,
    # else do nothing (keep the extreme values)
    if (x is not None) and (min_dynamic.sum() > 0):
        feature_min[min_dynamic] = x[min_dynamic]
    if (x is not None) and (max_dynamic.sum() > 0):
        feature_max[max_dynamic] = x[max_dynamic]

    # Raise warning if dynamic input waited but not provided
    dynamic_number = min_dynamic.sum() + max_dynamic.sum()
    if dynamic_number > 0 and x is None:
        logging.getLogger().warning(
            f"{dynamic_number} feature min and max are dynamic but no "
            "input were provided."
        )

    return feature_min, feature_max


def fix_feature_types(
    constraints: Constraints, x_clean: NDFloat, x_adv: NDFloat
) -> NDFloat:
    int_type_mask = constraints.feature_types != "real"
    if int_type_mask.sum() > 0:
        x_adv = x_adv.copy()
        x_adv_int = x_adv[..., int_type_mask]
        x_clean_int = x_clean[..., int_type_mask]
        x_plus_minus = (
            x_adv_int
            - np.repeat(
                x_clean_int[:, np.newaxis, :], x_adv_int.shape[1], axis=1
            )
        ) >= 0

        x_adv_int[x_plus_minus] = np.floor(x_adv_int[x_plus_minus])
        x_adv_int[~x_plus_minus] = np.ceil(x_adv_int[~x_plus_minus])
        x_adv[..., int_type_mask] = x_adv_int
    return x_adv



from math import sqrt
from typing import Any, List, Optional, Tuple, Type, Dict
from joblib import parallel_backend
import optuna
import pandas as pd

from sklearn.ensemble import RandomForestClassifier
from mlc.models.model import Model
from mlc.models.sk_models import SkModel
from mlc.transformers.tab_scaler import TabScaler
from mlc.typing import NDFloat, NDNumber


class Rf(SkModel):
    def __init__(
        self,
        name: str = "random_forest",
        objective: str = "classification",
        x_metadata: Optional[pd.DataFrame] = None,
        scaler: Optional[TabScaler] = None,
        n_jobs: int = -1,
        **kwargs: Any,
    ):
        super().__init__(
            name=name,
            objective=objective,
            x_metadata=x_metadata,
            scaler=scaler,
            n_jobs=n_jobs,
            **kwargs,
        )
        self.x_metadata = x_metadata
        self.scaler = scaler

        if self.scaler is not None:
            n_scaler = TabScaler(num_scaler="min_max", one_hot_encode=True)
            n_scaler.fit_scaler_data(self.scaler.get_scaler_data())
            self.scaler = n_scaler

        rf_parameters = {
            "class_weight": "balanced",
            **kwargs,
        }

        # Filter out parameters that are not in the model
        rf_parameters = {
            e: kwargs[e]
            for e in rf_parameters
            if e in RandomForestClassifier().get_params()
        }

        rf_parameters = {
            **rf_parameters,
            "n_jobs": n_jobs,
            "verbose": 1,
        }

        self.model = RandomForestClassifier(
            **rf_parameters,
        )

    @staticmethod
    def define_trial_parameters(
        trial: optuna.Trial, trial_params: Dict[str, Any]
    ) -> Dict[str, Any]:

        len_x = len(trial_params["x_metadata"])
        max_features = 1.0
        if len_x > 200:
            max_features = sqrt(len_x) / len_x

        params = {
            "n_estimators": trial.suggest_int("n_estimators", 10, 1000),
            "max_depth": trial.suggest_int("max_depth", 4, 50),
            "min_samples_split": trial.suggest_int(
                "min_samples_split", 2, 150
            ),
            "min_samples_leaf": trial.suggest_int("min_samples_leaf", 1, 60),
            "max_features": trial.suggest_float(
                "max_features", 0, max_features
            ),
        }

        return params

    @staticmethod
    def get_default_params(trial_params: Dict[str, Any]) -> Dict[str, Any]:
        params = {
            "n_estimators": 100,
            "max_depth": 50,
            "min_samples_split": 2,
            "min_samples_leaf": 1,
            "max_features": (
                sqrt(trial_params["n_features"]) / trial_params["n_features"]
            ),
        }

        return params

    def predict(self, x: NDFloat) -> NDNumber:
        """
        Returns the regression value or the concrete classes of binary /
        multi-class-classification tasks.
        (Save predictions to self.predictions)

        :param x: test data
        :return: predicted values / classes of test data (Shape N x 1)
        """
        if self.scaler is not None:
            x = self.scaler.transform(x)
        return super().predict(x)

    def predict_proba(self, x: NDFloat) -> NDFloat:
        """
        Only implemented for binary / multi-class-classification tasks.
        Returns the probability distribution over the classes C.
        (Save probabilities to self.prediction_probabilities)

        :param x: test data
        :return: probabilities for the classes (Shape N x C)
        """
        if self.scaler is not None:
            x = self.scaler.transform(x)
        return super().predict_proba(x)

    def fit(
        self,
        x: NDFloat,
        y: NDNumber,
        x_val: Optional[NDFloat] = None,
        y_val: Optional[NDNumber] = None,
    ) -> None:
        x = self.scaler.transform(x)
        if x_val is not None:
            x_val = self.scaler.transform(x_val)

        super().fit(
            x,
            y,
            x_val,
            y_val,
        )

    def set_experiment(self, experiment: Any) -> None:
        pass

    @staticmethod
    def load_class(self, path: str, **kwargs) -> None:
        model = Rf(**kwargs)
        model.load(path)
        return model


models: List[Tuple[str, Type[Model]]] = [("random_forest", Rf)]



import statistics
from dataclasses import dataclass
from typing import Any, Callable, Literal, cast

import rtdl
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import zero
from torch import Tensor

from .util import TaskType


def cos_sin(x: Tensor) -> Tensor:
    return torch.cat([torch.cos(x), torch.sin(x)], -1)


@dataclass
class PeriodicOptions:
    n: int  # the output size is 2 * n
    sigma: float
    trainable: bool
    initialization: Literal['log-linear', 'normal']


class Periodic(nn.Module):
    def __init__(self, n_features: int, options: PeriodicOptions) -> None:
        super().__init__()
        if options.initialization == 'log-linear':
            coefficients = options.sigma ** (torch.arange(options.n) / options.n)
            coefficients = coefficients[None].repeat(n_features, 1)
        else:
            assert options.initialization == 'normal'
            coefficients = torch.normal(0.0, options.sigma, (n_features, options.n))
        if options.trainable:
            self.coefficients = nn.Parameter(coefficients)  # type: ignore[code]
        else:
            self.register_buffer('coefficients', coefficients)

    def forward(self, x: Tensor) -> Tensor:
        assert x.ndim == 2
        return cos_sin(2 * torch.pi * self.coefficients[None] * x[..., None])


def get_n_parameters(m: nn.Module):
    return sum(x.numel() for x in m.parameters() if x.requires_grad)


def get_loss_fn(task_type: TaskType) -> Callable[..., Tensor]:
    return (
        F.binary_cross_entropy_with_logits
        if task_type == TaskType.BINCLASS
        else F.cross_entropy
        if task_type == TaskType.MULTICLASS
        else F.mse_loss
    )


def default_zero_weight_decay_condition(module_name, module, parameter_name, parameter):
    del module_name, parameter
    return parameter_name.endswith('bias') or isinstance(
        module,
        (
            nn.BatchNorm1d,
            nn.LayerNorm,
            nn.InstanceNorm1d,
            rtdl.CLSToken,
            rtdl.NumericalFeatureTokenizer,
            rtdl.CategoricalFeatureTokenizer,
            Periodic,
        ),
    )


def split_parameters_by_weight_decay(
        model: nn.Module, zero_weight_decay_condition=default_zero_weight_decay_condition
) -> list[dict[str, Any]]:
    parameters_info = {}
    for module_name, module in model.named_modules():
        for parameter_name, parameter in module.named_parameters():
            full_parameter_name = (
                f'{module_name}.{parameter_name}' if module_name else parameter_name
            )
            parameters_info.setdefault(full_parameter_name, ([], parameter))[0].append(
                zero_weight_decay_condition(
                    module_name, module, parameter_name, parameter
                )
            )
    params_with_wd = {'params': []}
    params_without_wd = {'params': [], 'weight_decay': 0.0}
    for full_parameter_name, (results, parameter) in parameters_info.items():
        (params_without_wd if any(results) else params_with_wd)['params'].append(
            parameter
        )
    return [params_with_wd, params_without_wd]


def make_optimizer(
        config: dict[str, Any],
        parameter_groups,
) -> optim.Optimizer:
    if config['optimizer'] == 'FT-Transformer-default':
        return optim.AdamW(parameter_groups, lr=1e-4, weight_decay=1e-5)
    return getattr(optim, config['optimizer'])(
        parameter_groups,
        **{x: config[x] for x in ['lr', 'weight_decay', 'momentum'] if x in config},
    )


def get_lr(optimizer: optim.Optimizer) -> float:
    return next(iter(optimizer.param_groups))['lr']


def is_oom_exception(err: RuntimeError) -> bool:
    return any(
        x in str(err)
        for x in [
            'CUDA out of memory',
            'CUBLAS_STATUS_ALLOC_FAILED',
            'CUDA error: out of memory',
        ]
    )


def train_with_auto_virtual_batch(
        optimizer,
        loss_fn,
        step,
        batch,
        chunk_size: int,
) -> tuple[Tensor, int]:
    batch_size = len(batch)
    random_state = zero.random.get_state()
    loss = None
    while chunk_size != 0:
        try:
            zero.random.set_state(random_state)
            optimizer.zero_grad()
            if batch_size <= chunk_size:
                loss = loss_fn(*step(batch))
                loss.backward()
            else:
                loss = None
                for chunk in zero.iter_batches(batch, chunk_size):
                    chunk_loss = loss_fn(*step(chunk))
                    chunk_loss = chunk_loss * (len(chunk) / batch_size)
                    chunk_loss.backward()
                    if loss is None:
                        loss = chunk_loss.detach()
                    else:
                        loss += chunk_loss.detach()
        except RuntimeError as err:
            if not is_oom_exception(err):
                raise
            chunk_size //= 2
        else:
            break
    if not chunk_size:
        raise RuntimeError('Not enough memory even for batch_size=1')
    optimizer.step()
    return cast(Tensor, loss), chunk_size


def process_epoch_losses(losses: list[Tensor]) -> tuple[list[float], float]:
    losses_ = torch.stack(losses).tolist()
    return losses_, statistics.mean(losses_)



import os
import pickle
from typing import Tuple

import numpy as np
import torch
from torch.utils.data import DataLoader, TensorDataset


def get_semisup_dataloaders(
        train_dataset,
        test_dataset,
        val_dataset=None,
        batch_size=256,
        batch_size_test=256,
        num_workers=4,
        unsup_fraction=0.5,
):
    """
    Return dataloaders with custom sampling of pseudo-labeled data.
    """
    dataset_size = train_dataset.dataset_size
    train_batch_sampler = SemiSupervisedSampler(
        train_dataset.sup_indices,
        train_dataset.unsup_indices,
        batch_size,
        unsup_fraction,
        num_batches=int(np.ceil(dataset_size / batch_size)),
    )
    # epoch_size = len(train_batch_sampler) * batch_size

    kwargs = {
        "num_workers": num_workers,
        "pin_memory": torch.cuda.is_available(),
    }
    train_dataloader = torch.utils.data.DataLoader(
        train_dataset, batch_sampler=train_batch_sampler, **kwargs
    )
    test_dataloader = torch.utils.data.DataLoader(
        test_dataset, batch_size=batch_size_test, shuffle=False, **kwargs
    )

    if val_dataset:
        val_dataloader = torch.utils.data.DataLoader(
            val_dataset, batch_size=batch_size_test, shuffle=False, **kwargs
        )
        return train_dataloader, test_dataloader, val_dataloader
    return train_dataloader, test_dataloader


class SemiSupervisedDataset(torch.utils.data.Dataset):
    """
    A dataset with auxiliary pseudo-labeled data.
    """

    def __init__(
            self,
            x: torch.Tensor,
            y: torch.Tensor,
            take_amount=None,
            take_amount_seed=13,
            aux_data_filename=None,
            add_aux_labels=False,
            aux_take_amount=None,
            train=False,
            **kwargs,
    ):

        self.train = train

        self.dataset = TensorDataset(x, y)

        self.dataset_size = len(self.dataset)

        if self.train:
            if take_amount is not None:
                rng_state = np.random.get_state()
                np.random.seed(take_amount_seed)
                take_inds = np.random.choice(
                    len(self.sup_indices), take_amount, replace=False
                )
                np.random.set_state(rng_state)

                self.targets = self.targets[take_inds]
                self.data = self.data[take_inds]

            self.sup_indices = list(range(len(self.targets)))
            self.unsup_indices = []

            if aux_data_filename is not None:
                aux_path = aux_data_filename
                print("Loading data from %s" % aux_path)
                if os.path.splitext(aux_path)[1] == ".pickle":
                    # for data from Carmon et al, 2019.
                    with open(aux_path, "rb") as f:
                        aux = pickle.load(f)
                    aux_data = aux["data"]
                    aux_targets = aux["extrapolated_targets"]
                else:
                    # for data from Rebuffi et al, 2021.
                    aux = np.load(aux_path)
                    aux_data = aux["image"]
                    aux_targets = aux["label"]

                orig_len = len(self.data)

                if aux_take_amount is not None:
                    rng_state = np.random.get_state()
                    np.random.seed(take_amount_seed)
                    take_inds = np.random.choice(
                        len(aux_data), aux_take_amount, replace=False
                    )
                    np.random.set_state(rng_state)

                    aux_data = aux_data[take_inds]
                    aux_targets = aux_targets[take_inds]

                self.data = np.concatenate((self.data, aux_data), axis=0)

                if not add_aux_labels:
                    self.targets.extend([-1] * len(aux_data))
                else:
                    self.targets.extend(aux_targets)
                self.unsup_indices.extend(
                    range(orig_len, orig_len + len(aux_data))
                )

        else:
            self.sup_indices = list(range(len(self.targets)))
            self.unsup_indices = []

    def load_base_dataset(self, **kwargs):
        raise NotImplementedError()

    @property
    def data(self):
        return self.dataset.tensors[0]

    @data.setter
    def data(self, value):
        self.dataset.tensors[0] = value

    @property
    def targets(self):
        return self.dataset.tensors[1]

    @targets.setter
    def targets(self, value):
        self.dataset.tensors[1] = value

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, item):
        self.dataset.labels = self.targets
        return self.dataset[item]


class SemiSupervisedSampler(torch.utils.data.Sampler):
    """
    Balanced sampling from the labeled and unlabeled data.
    """

    def __init__(
            self,
            sup_inds,
            unsup_inds,
            batch_size,
            unsup_fraction=0.5,
            num_batches=None,
    ):
        if unsup_fraction is None or unsup_fraction < 0:
            self.sup_inds = sup_inds + unsup_inds
            unsup_fraction = 0.0
        else:
            self.sup_inds = sup_inds
            self.unsup_inds = unsup_inds

        self.batch_size = batch_size
        unsup_batch_size = int(batch_size * unsup_fraction)
        self.sup_batch_size = batch_size - unsup_batch_size

        if num_batches is not None:
            self.num_batches = num_batches
        else:
            self.num_batches = int(
                np.ceil(len(self.sup_inds) / self.sup_batch_size)
            )
        super().__init__(None)

    def __iter__(self):
        batch_counter = 0
        while batch_counter < self.num_batches:
            sup_inds_shuffled = [
                self.sup_inds[i] for i in torch.randperm(len(self.sup_inds))
            ]
            for sup_k in range(0, len(self.sup_inds), self.sup_batch_size):
                if batch_counter == self.num_batches:
                    break
                batch = sup_inds_shuffled[
                        sup_k: (sup_k + self.sup_batch_size)
                        ]
                if self.sup_batch_size < self.batch_size:
                    batch.extend(
                        [
                            self.unsup_inds[i]
                            for i in torch.randint(
                            high=len(self.unsup_inds),
                            size=(self.batch_size - len(batch),),
                            dtype=torch.int64,
                        )
                        ]
                    )
                np.random.shuffle(batch)
                yield batch
                batch_counter += 1

    def __len__(self):
        return self.num_batches


class SemiSupervisedDataLoader:
    def __init__(
            self,
            take_amount=None,
            unsup_fraction: float = 0.5,
            num_workers: int = 4,
    ) -> None:
        self.take_amount = take_amount
        self.unsup_fraction = unsup_fraction
        self.num_workers = num_workers

    def get_dataloader(
            self, x: torch.Tensor, y: torch.Tensor, train: bool, batch_size: int
    ) -> DataLoader[Tuple[torch.Tensor, ...]]:

        # Create dataset
        dataset = SemiSupervisedDataset(
            x,
            y,
            take_amount=self.take_amount,
            take_amount_seed=13,
            aux_data_filename=None,
            add_aux_labels=False,
            aux_take_amount=None,
            train=train,
        )

        # Create data sampler

        if train:
            dataset_size = dataset.dataset_size
            train_batch_sampler = SemiSupervisedSampler(
                dataset.sup_indices,
                dataset.unsup_indices,
                batch_size,
                self.unsup_fraction,
                num_batches=int(np.ceil(dataset_size / batch_size)),
            )

        # Create dataloader

        kwargs = {
            "num_workers": self.num_workers,
        }
        if train:
            return torch.utils.data.DataLoader(
                dataset, batch_sampler=train_batch_sampler, **kwargs
            )

        else:
            return torch.utils.data.DataLoader(
                dataset, batch_size=batch_size, shuffle=False, **kwargs
            )



from typing import Any

import numpy as np
import numpy.typing as npt

from mlc.constraints.constraints import Constraints, get_feature_min_max
from mlc.constraints.constraints_backend_executor import ConstraintsExecutor
from mlc.constraints.numpy_backend import NumpyBackend
from mlc.constraints.relation_constraint import AndConstraint
from mlc.typing import NDBool, NDFloat


class ConstraintChecker:
    def __init__(self, constraints: Constraints, tolerance: float = 0.0):
        self.constraints = constraints
        self.tolerance = tolerance

    def _check_relationship_constraints(
        self, x_adv: npt.NDArray[Any]
    ) -> NDBool:

        constraints_executor = ConstraintsExecutor(
            AndConstraint(self.constraints.relation_constraints),
            backend=NumpyBackend(),
            feature_names=self.constraints.feature_names,
        )
        out = constraints_executor.execute(x_adv)
        if not isinstance(out, np.ndarray):
            raise ValueError(
                "ConstraintExecutor did not return a numpy array."
            )
        return out <= self.tolerance

    def _check_boundary_constraints(
        self, x: NDFloat, x_adv: NDFloat
    ) -> NDBool:
        xl, xu = get_feature_min_max(self.constraints, x)
        xl_ok, xu_ok = np.min(
            (xl - np.finfo(np.float32).eps) <= x_adv, axis=1
        ), np.min((xu + np.finfo(np.float32).eps) >= x_adv, axis=1)
        return xl_ok * xu_ok

    def _check_type_constraints(self, x_adv: NDFloat) -> NDBool:
        int_type_mask = self.constraints.feature_types != "real"
        if int_type_mask.sum() > 0:
            type_ok = np.min(
                (x_adv[:, int_type_mask] == np.round(x_adv[:, int_type_mask])),
                axis=1,
            )
        else:
            type_ok = np.ones(shape=x_adv.shape[:-1], dtype=np.bool_)
        return type_ok

    def _check_mutable_constraints(self, x: NDFloat, x_adv: NDFloat) -> NDBool:
        immutable_mask = ~self.constraints.mutable_features
        if immutable_mask.sum() > 0:
            mutable_ok = np.min(
                (x[:, immutable_mask] == x_adv[:, immutable_mask]), axis=1
            )
        else:
            mutable_ok = np.ones(shape=x_adv.shape[:-1], dtype=np.bool_)
        return mutable_ok

    def check_constraints(self, x: NDFloat, x_adv: NDFloat) -> NDFloat:
        constraints = np.array(
            [
                self._check_relationship_constraints(x_adv),
                self._check_boundary_constraints(x, x_adv),
                self._check_type_constraints(x_adv),
                self._check_mutable_constraints(x, x_adv),
            ]
        )
        constraints = np.min(constraints, axis=0)
        return constraints



from turtle import pd
from typing import Any, Dict, Optional, Union
import numpy as np
import optuna
import torch
import torch.nn as nn
import torch.nn.functional as F

from mlc.models.torch_models import BaseModelTorch
from mlc.transformers.tab_scaler import TabScaler
from mlc.typing import NDFloat, NDNumber

"""
    Custom implementation for the standard multi-layer perceptron
"""


class MLP(BaseModelTorch):
    def __init__(
        self,
        objective: str,
        n_layers: int,
        input_dim: int,
        hidden_dim: int,
        output_dim: int,
        scaler: Optional[TabScaler] = None,
        **kwargs: Any,
    ):

        self.objective = objective
        self.n_layers = n_layers
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.output_dim = output_dim
        self.scaler = scaler

        self.model = MLPModel(
            n_layers=self.n_layers,
            input_dim=self.input_dim,
            hidden_dim=self.hidden_dim,
            output_dim=self.output_dim,
            task=self.objective,
        )

        if self.scaler is not None:
            self.model = nn.Sequential(
                self.scaler.get_transorm_nn(), self.model
            )

        super().__init__(
            objective=objective,
            n_layers=n_layers,
            input_dim=input_dim,
            hidden_dim=hidden_dim,
            output_dim=output_dim,
            scaler=scaler,
            **kwargs,
        )

        self.to_device()

    def fit(
        self,
        x: NDFloat,
        y: NDNumber,
        x_val: Optional[NDFloat] = None,
        y_val: Optional[NDNumber] = None,
    ):
        x = np.array(x, dtype=np.float)
        x_val = np.array(x_val, dtype=np.float)

        return super().fit(x, y, x_val, y_val)

    def predict_helper(
        self,
        x: Union[NDFloat, torch.Tensor, pd.DataFrame],
        load_all_gpu: bool = False,
    ):
        x = np.array(x, dtype=np.float)
        return super().predict_helper(x, load_all_gpu)

    @staticmethod
    def define_trial_parameters(
        trial: optuna.Trial, trial_params: Dict[str, Any]
    ):
        params = {
            "hidden_dim": trial.suggest_int("hidden_dim", 10, 100),
            "n_layers": trial.suggest_int("n_layers", 2, 5),
            "learning_rate": trial.suggest_float(
                "learning_rate", 0.0005, 0.001
            ),
        }
        return params


class MLPModel(nn.Module):
    def __init__(self, n_layers, input_dim, hidden_dim, output_dim, task):
        super().__init__()

        self.task = task

        self.layers = nn.ModuleList()

        # Input Layer (= first hidden layer)
        self.input_layer = nn.Linear(input_dim, hidden_dim)

        # Hidden Layers (number specified by n_layers)
        self.layers.extend(
            [nn.Linear(hidden_dim, hidden_dim) for _ in range(n_layers - 1)]
        )

        # Output Layer
        self.output_layer = nn.Linear(hidden_dim, output_dim)

    def forward(self, x):
        x = F.relu(self.input_layer(x))

        # Use ReLU as activation for all hidden layers
        for layer in self.layers:
            x = F.relu(layer(x))

        # No activation function on the output
        x = self.output_layer(x)

        if self.task == "classification":
            x = F.softmax(x, dim=1)

        return x



import numpy as np
from sklearn.metrics import (
    classification_report,
    f1_score,
    matthews_corrcoef,
    precision_score,
    recall_score,
    roc_auc_score,
)


def compute_binary_metrics(y_true, y_score, threshold=None) -> dict:
    if threshold is None:
        y_pred = np.argmax(y_score, axis=1)
    else:
        y_pred = (y_score[:, 1] >= threshold).astype(int)

    metrics = {
        **classification_report(y_true, y_pred, output_dict=True),
        **{
            "roc_auc_score": roc_auc_score(y_true, y_score[:, 1]),
            "precision_score": precision_score(y_true, y_pred),
            "recall_score": recall_score(y_true, y_pred),
            "f1_score": f1_score(y_true, y_pred),
            "matthews_corrcoef": matthews_corrcoef(y_true, y_pred),
        },
    }
    return metrics



## Adapted from https://github.com/yandex-research/tab-ddpm/
from typing import Tuple

import torch
from torch.utils.data import DataLoader, TensorDataset

from mlc.dataloaders.dataloader import BaseDataLoader
from mlc.dataloaders.generative.tab_ddpm.lib.data import Dataset
from mlc.dataloaders.generative.tab_ddpm.lib.util import TaskType
from mlc.dataloaders.generative.tab_ddpm.sample import sample
from mlc.dataloaders.generative.tab_ddpm.train import train, get_model
from mlc.datasets.dataset import Dataset as MLC_Dataset


class DDPMDataLoader(BaseDataLoader):
    def __init__(self, num_workers: int = 2, original_dataset: MLC_Dataset = None, ddpm_params: dict = {},
                 pretrained_path: str = None, save_path: str = None, nb_classes: int = 2,
                 task_type: str = TaskType.MULTICLASS, train_batch_size: int = 32) -> None:

        self.num_workers = num_workers
        task = TaskType.BINCLASS if nb_classes == 2 else task_type
        if pretrained_path is None:

            x, y = original_dataset.get_x_y()
            metadata = original_dataset.get_metadata(only_x=True)
            transformations = original_dataset.get_ddpm_transformations()
            num_index = [k for (k, v) in enumerate(metadata.type.values) if v != "cat"]
            cat_index = [k for (k, v) in enumerate(metadata.type.values) if v == "cat"]
            dataset = Dataset({"train": x.values[:, num_index]}, {"train": x.values[:, cat_index]}, {"train": y}, {},
                              task,
                              nb_classes)

            self.generator = train("../", dataset=dataset, T_dict=transformations, batch_size=train_batch_size,
                                   **ddpm_params)
        else:
            self.generator = get_model(
                ddpm_params.get("model_type"),
                ddpm_params.get("model_params"),
                ddpm_params.get("num_numerical_features_"),
                category_sizes=dataset.get_category_sizes('train')
            )

            self.generator.load_state_dict(
                torch.load(pretrained_path, map_location="cpu")
            )

        if save_path is not None:
            self.save(save_path)

    def load(self, pretrained_path: str):
        self.generator._model.load(pretrained_path)

    def save(self, save_path: str):
        self.generator._model.save(save_path)

    def get_dataloader(
            self, x: torch.Tensor, y: torch.Tensor, train: bool, batch_size: int, augment: bool = True
    ) -> DataLoader[Tuple[torch.Tensor, ...]]:

        # Create dataset
        x_aug, y_aug = sample("../", model=self.generator, num_samples=len(y), batch_size=batch_size)

        x_aug, y_aug = torch.Tensor(x_aug).type(x), torch.Tensor(y_aug).type(y)
        if augment:
            dataset = TensorDataset(torch.cat((x, x_aug)), torch.cat((y, y_aug)))
        else:
            dataset = TensorDataset(x_aug, y_aug)

        # Create dataloader
        if train:

            return DataLoader(
                dataset=dataset,
                batch_size=batch_size,
                shuffle=True,
                num_workers=self.num_workers,
            )

        else:
            return DataLoader(
                dataset=dataset,
                batch_size=batch_size,
                shuffle=False,
                num_workers=self.num_workers,
            )



import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from .activation import activation_layer


class LocalActivationUnit(nn.Module):
    """The LocalActivationUnit used in DIN with which the representation of
        user interests varies adaptively given different candidate items.

    Input shape
        - A list of two 3D tensor with shape:  ``(batch_size, 1, embedding_size)`` and ``(batch_size, T, embedding_size)``

    Output shape
        - 3D tensor with shape: ``(batch_size, T, 1)``.

    Arguments
        - **hidden_units**:list of positive integer, the attention net layer number and units in each layer.

        - **activation**: Activation function to use in attention net.

        - **l2_reg**: float between 0 and 1. L2 regularizer strength applied to the kernel weights matrix of attention net.

        - **dropout_rate**: float in [0,1). Fraction of the units to dropout in attention net.

        - **use_bn**: bool. Whether use BatchNormalization before activation or not in attention net.

        - **seed**: A Python integer to use as random seed.

    References
        - [Zhou G, Zhu X, Song C, et al. Deep interest network for click-through rate prediction[C]//Proceedings of the 24th ACM SIGKDD International Conference on Knowledge Discovery & Data Mining. ACM, 2018: 1059-1068.](https://arxiv.org/pdf/1706.06978.pdf)
    """

    def __init__(self, hidden_units=(64, 32), embedding_dim=4, activation='sigmoid', dropout_rate=0, dice_dim=3,
                 l2_reg=0, use_bn=False):
        super(LocalActivationUnit, self).__init__()

        self.dnn = DNN(inputs_dim=4 * embedding_dim,
                       hidden_units=hidden_units,
                       activation=activation,
                       l2_reg=l2_reg,
                       dropout_rate=dropout_rate,
                       dice_dim=dice_dim,
                       use_bn=use_bn)

        self.dense = nn.Linear(hidden_units[-1], 1)

    def forward(self, query, user_behavior):
        # query ad            : size -> batch_size * 1 * embedding_size
        # user behavior       : size -> batch_size * time_seq_len * embedding_size
        user_behavior_len = user_behavior.size(1)

        queries = query.expand(-1, user_behavior_len, -1)

        attention_input = torch.cat([queries, user_behavior, queries - user_behavior, queries * user_behavior],
                                    dim=-1)  # as the source code, subtraction simulates verctors' difference
        attention_output = self.dnn(attention_input)

        attention_score = self.dense(attention_output)  # [B, T, 1]

        return attention_score


class DNN(nn.Module):
    """The Multi Layer Percetron

      Input shape
        - nD tensor with shape: ``(batch_size, ..., input_dim)``. The most common situation would be a 2D input with shape ``(batch_size, input_dim)``.

      Output shape
        - nD tensor with shape: ``(batch_size, ..., hidden_size[-1])``. For instance, for a 2D input with shape ``(batch_size, input_dim)``, the output would have shape ``(batch_size, hidden_size[-1])``.

      Arguments
        - **inputs_dim**: input feature dimension.

        - **hidden_units**:list of positive integer, the layer number and units in each layer.

        - **activation**: Activation function to use.

        - **l2_reg**: float between 0 and 1. L2 regularizer strength applied to the kernel weights matrix.

        - **dropout_rate**: float in [0,1). Fraction of the units to dropout.

        - **use_bn**: bool. Whether use BatchNormalization before activation or not.

        - **seed**: A Python integer to use as random seed.
    """

    def __init__(self, inputs_dim, hidden_units, activation='relu', l2_reg=0, dropout_rate=0, use_bn=False,
                 init_std=0.0001, dice_dim=3, seed=1024, device='cpu'):
        super(DNN, self).__init__()
        self.dropout_rate = dropout_rate
        self.dropout = nn.Dropout(dropout_rate)
        self.seed = seed
        self.l2_reg = l2_reg
        self.use_bn = use_bn
        if len(hidden_units) == 0:
            raise ValueError("hidden_units is empty!!")
        hidden_units = [inputs_dim] + list(hidden_units)

        self.linears = nn.ModuleList(
            [nn.Linear(hidden_units[i], hidden_units[i + 1]) for i in range(len(hidden_units) - 1)])

        if self.use_bn:
            self.bn = nn.ModuleList(
                [nn.BatchNorm1d(hidden_units[i + 1]) for i in range(len(hidden_units) - 1)])

        self.activation_layers = nn.ModuleList(
            [activation_layer(activation, hidden_units[i + 1], dice_dim) for i in range(len(hidden_units) - 1)])

        for name, tensor in self.linears.named_parameters():
            if 'weight' in name:
                nn.init.normal_(tensor, mean=0, std=init_std)

        self.to(device)

    def forward(self, inputs):
        deep_input = inputs

        for i in range(len(self.linears)):

            fc = self.linears[i](deep_input)

            if self.use_bn:
                fc = self.bn[i](fc)

            fc = self.activation_layers[i](fc)

            fc = self.dropout(fc)
            deep_input = fc
        return deep_input


class PredictionLayer(nn.Module):
    """
      Arguments
         - **task**: str, ``"binary"`` for  binary logloss or  ``"regression"`` for regression loss
         - **use_bias**: bool.Whether add bias term or not.
    """

    def __init__(self, task='binary', use_bias=True, **kwargs):
        if task not in ["binary", "multiclass", "regression"]:
            raise ValueError("task must be binary,multiclass or regression")

        super(PredictionLayer, self).__init__()
        self.use_bias = use_bias
        self.task = task
        if self.use_bias:
            self.bias = nn.Parameter(torch.zeros((1,)))

    def forward(self, X):
        output = X
        if self.use_bias:
            output += self.bias
        if self.task == "binary":
            output = torch.sigmoid(output)
        return output


class Conv2dSame(nn.Conv2d):
    """ Tensorflow like 'SAME' convolution wrapper for 2D convolutions
    """

    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True):
        super(Conv2dSame, self).__init__(
            in_channels, out_channels, kernel_size, stride, 0, dilation,
            groups, bias)
        nn.init.xavier_uniform_(self.weight)

    def forward(self, x):
        ih, iw = x.size()[-2:]
        kh, kw = self.weight.size()[-2:]
        oh = math.ceil(ih / self.stride[0])
        ow = math.ceil(iw / self.stride[1])
        pad_h = max((oh - 1) * self.stride[0] + (kh - 1) * self.dilation[0] + 1 - ih, 0)
        pad_w = max((ow - 1) * self.stride[1] + (kw - 1) * self.dilation[1] + 1 - iw, 0)
        if pad_h > 0 or pad_w > 0:
            x = F.pad(x, [pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2])
        out = F.conv2d(x, self.weight, self.bias, self.stride,
                       self.padding, self.dilation, self.groups)
        return out



# The SAINT model.
from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Type

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from einops import rearrange
from torch import einsum
from torch.utils.data import DataLoader, TensorDataset

from mlc.load_do_save import load_json, save_json
from mlc.models.model import Model
from mlc.models.tabsurvey.saint_lib.augmentations import embed_data_mask
from mlc.models.tabsurvey.saint_lib.data_openml import DataSetCatCon
from mlc.models.tabsurvey.saint_lib.models.pretrainmodel import (
    SAINT as SAINTModel,
)
from mlc.models.torch_models import BaseModelTorch
from mlc.transformers.tab_scaler import TabScaler
from mlc.utils import parent_exists, to_torch_number

"""
    SAINT: Improved Neural Networks for Tabular Data via
    Row Attention and Contrastive Pre-Training
    (https://arxiv.org/abs/2106.01342)

    Code adapted from: https://github.com/somepago/saint
"""
import optuna


class SAINTWrapper(nn.Module):
    def __init__(self, cat_idx, num_idx, task, model, device) -> None:
        super().__init__()
        self.cat_idx = cat_idx
        self.num_idx = num_idx
        self.task = task
        self.model = model
        self.device = device

    def forward(self, X: torch.Tensor) -> torch.Tensor:

        device = self.device
        Y = torch.ones((X.shape[0], 1))
        X_mask = torch.ones_like(X)

        X = torch.clone(X)

        self.X1 = (
            X[:, self.cat_idx].clone().type(torch.int64)
        )  # categorical columns
        self.X2 = (
            X[:, self.num_idx].clone().type(torch.float32)
        )  # numerical columns
        self.X1_mask = (
            X_mask[:, self.cat_idx].clone().type(torch.int64)
        )  # categorical columns
        self.X2_mask = (
            X_mask[:, self.num_idx].clone().type(torch.int64)
        )  # numerical columns
        if self.task == "regression":
            self.y = Y.astype(np.float32)
        else:
            self.y = Y  # .astype(np.float32)
        self.cls = torch.zeros_like(self.y, dtype=int)
        self.cls_mask = torch.ones_like(self.y, dtype=int)

        x_categ, x_cont, cat_mask, con_mask = (
            torch.concat((self.cls, self.X1), dim=1),
            self.X2,
            torch.concat((self.cls_mask, self.X1_mask), dim=1),
            self.X2_mask,
        )

        x_categ, x_cont, cat_mask, con_mask = (
            x_categ.to(device),
            x_cont.to(device),
            cat_mask.to(device),
            con_mask.to(device),
        )

        _, x_categ_enc, x_cont_enc = embed_data_mask(
            x_categ, x_cont, cat_mask, con_mask, self.model
        )
        reps = self.model.transformer(x_categ_enc, x_cont_enc)
        y_reps = reps[:, 0, :]
        y_outs = self.model.mlpfory(y_reps)

        return y_outs


class SAINT(BaseModelTorch):
    def __init__(
        self,
        objective: str,
        x_metadata: pd.DataFrame,
        batch_size: int,
        epochs: int,
        early_stopping_rounds: int,
        num_classes: int,
        dim: int = 64,
        depth: int = 6,
        heads: int = 8,
        dropout: float = 0.1,
        val_batch_size: int = 100000,
        name="saint",
        scaler: Optional[TabScaler] = None,
        **kwargs,
    ):

        # Thibault: this should be handled somewhere else
        # Decreasing some hyperparameter to cope with memory issues
        # dim = self.params["dim"] if args.num_features < 20000 else 8
        # self.batch_size = self.args.batch_size
        # depth = self.params["depth"] if args.num_features < 20000 else 3
        # heads = self.params["heads"] if args.num_features < 20000 else 4
        # print("Using dim %d and batch size %d" % (dim, self.batch_size))

        self.objective = objective
        self.x_metadata = x_metadata
        self.batch_size = batch_size
        self.epochs = epochs
        self.early_stopping_rounds = early_stopping_rounds
        self.num_classes = num_classes

        self.dim = dim
        self.depth = depth
        self.heads = heads
        self.dropout = dropout
        self.val_batch_size = val_batch_size

        if scaler is not None:
            self.scaler = TabScaler(num_scaler="min_max", one_hot_encode=False)
            self.scaler.fit_scaler_data(scaler.get_scaler_data())

        # Generate super call
        super().__init__(
            objective=objective,
            x_metadata=x_metadata,
            batch_size=batch_size,
            epochs=epochs,
            early_stopping_rounds=early_stopping_rounds,
            num_classes=num_classes,
            dim=dim,
            depth=depth,
            heads=heads,
            dropout=dropout,
            val_batch_size=val_batch_size,
            name=name,
            scaler=scaler,
            **kwargs,
        )

        # Generated

        self.cat_idx = np.where(x_metadata["type"] == "cat")[0].tolist()
        self.num_features = x_metadata.shape[0]
        self.cat_idx = np.where(x_metadata["type"] == "cat")[0].tolist()
        self.num_idx = list(set(range(self.num_features)) - set(self.cat_idx))
        self.cat_dims = [
            (int(x_metadata.iloc[i]["max"]) + 1) for i in self.cat_idx
        ]

        self.model = SAINTModel(
            categories=tuple(
                np.append(np.array([1]), np.array(self.cat_dims)).astype(int)
            ),
            num_continuous=len(self.num_idx),
            dim=dim,
            dim_out=1,
            depth=depth,  # 6
            heads=heads,  # 8
            attn_dropout=self.dropout,  # 0.1
            ff_dropout=self.dropout,  # 0.1
            mlp_hidden_mults=(4, 2),
            cont_embeddings="MLP",
            attentiontype="colrow",
            final_mlp_style="sep",
            y_dim=self.num_classes,
        )

        # if self.args.data_parallel:
        #     self.model.transformer = nn.DataParallel(
        #         self.model.transformer, device_ids=self.args.gpu_ids
        #     )
        #     self.model.mlpfory = nn.DataParallel(
        #         self.model.mlpfory, device_ids=self.args.gpu_ids
        #     )

        self.wrapper_model = SAINTWrapper(
            self.cat_idx, self.num_idx, self.objective, self.model, self.device
        )

        if hasattr(self, "scaler") and (self.scaler is not None):
            self.wrapper_model = nn.Sequential(
                self.scaler.get_transorm_nn(), self.wrapper_model
            )

        self.experiment = None

    def fit(self, X, y, X_val=None, y_val=None):

        if self.scaler is not None:
            X = self.scaler.transform(X)
            X_val = self.scaler.transform(X_val)
        self.set_loss_y(to_torch_number(y))
        criterion = self.loss_func

        # if (y.ndim == 1) or (y.shape[1] == 1):
        #     y = np.column_stack((1 - y, y))
        # if (y_val.ndim == 1) or (y_val.shape[1] == 1):
        #     y_val = np.column_stack((1 - y_val, y_val))

        optimizer = optim.AdamW(self.model.parameters(), lr=0.00003)

        self.model.to(self.device)

        # SAINT wants it like this...
        X = {"data": X, "mask": np.ones_like(X)}
        y = {"data": y.reshape(-1, 1)}
        X_val = {"data": X_val, "mask": np.ones_like(X_val)}
        y_val = {"data": y_val.reshape(-1, 1)}

        train_ds = DataSetCatCon(X, y, self.cat_idx, self.objective)
        trainloader = DataLoader(
            train_ds, batch_size=self.batch_size, shuffle=True, num_workers=4
        )

        val_ds = DataSetCatCon(X_val, y_val, self.cat_idx, self.objective)
        valloader = DataLoader(
            val_ds,
            batch_size=self.val_batch_size,
            shuffle=True,
            num_workers=4,
        )

        min_val_loss = float("inf")
        min_val_loss_idx = 0

        loss_history = []
        val_loss_history = []

        for epoch in range(self.epochs):
            self.model.train()

            for i, data in enumerate(trainloader, 0):
                print(
                    f"Epoch: {epoch} / {self.epochs}, "
                    f"Batch: {i} / {len(trainloader)}"
                )

                optimizer.zero_grad()

                # x_categ is the the categorical data,
                # x_cont has continuous data,
                # y_gts has ground truth ys.
                # cat_mask is an array of ones same shape as
                # x_categ and an additional column(corresponding to CLS
                # token) set to 0s.
                # con_mask is an array of ones same shape as x_cont.
                x_categ, x_cont, y_gts, cat_mask, con_mask = data

                x_categ, x_cont = x_categ.to(self.device), x_cont.to(
                    self.device
                )
                cat_mask, con_mask = cat_mask.to(self.device), con_mask.to(
                    self.device
                )

                # We are converting the data to embeddings in the next step
                _, x_categ_enc, x_cont_enc = embed_data_mask(
                    x_categ, x_cont, cat_mask, con_mask, self.model
                )

                reps = self.model.transformer(x_categ_enc, x_cont_enc)

                # select only the representations corresponding to CLS token
                # and apply mlp on it in the next step to get the predictions.
                y_reps = reps[:, 0, :]

                y_outs = self.model.mlpfory(y_reps)

                if self.objective == "regression":
                    y_gts = y_gts.to(self.device)
                elif self.objective == "classification":
                    y_gts = y_gts.to(self.device).squeeze()
                else:
                    y_gts = y_gts.to(self.device).float()

                loss = criterion(y_outs, y_gts)
                loss.backward()
                optimizer.step()

                loss_history.append(loss.item())
                if self.experiment is not None:
                    self.experiment.log_metric("train_loss", loss.item())

                print("Loss", loss.item())

            # Early Stopping
            val_loss = 0.0
            val_dim = 0
            self.model.eval()
            with torch.no_grad():
                for data in valloader:
                    x_categ, x_cont, y_gts, cat_mask, con_mask = data

                    x_categ, x_cont = x_categ.to(self.device), x_cont.to(
                        self.device
                    )
                    cat_mask, con_mask = cat_mask.to(self.device), con_mask.to(
                        self.device
                    )

                    _, x_categ_enc, x_cont_enc = embed_data_mask(
                        x_categ, x_cont, cat_mask, con_mask, self.model
                    )
                    reps = self.model.transformer(x_categ_enc, x_cont_enc)
                    y_reps = reps[:, 0, :]
                    y_outs = self.model.mlpfory(y_reps)

                    if self.objective == "regression":
                        y_gts = y_gts.to(self.device)
                    elif self.objective == "classification":
                        y_gts = y_gts.to(self.device).squeeze()
                    else:
                        y_gts = y_gts.to(self.device).float()

                    val_loss += criterion(y_outs, y_gts).item()
                    val_dim += 1
            val_loss /= val_dim

            val_loss_history.append(val_loss)
            if self.experiment is not None:
                self.experiment.log_metric("validation_loss", val_loss)

            print("Epoch", epoch, "loss", val_loss)

            if val_loss < min_val_loss:
                min_val_loss = val_loss
                min_val_loss_idx = epoch

                # Save the currently best model
                self.save_best_weights()
                # self.save_model(filename_extension="best", directory="tmp")

            if min_val_loss_idx + self.early_stopping_rounds < epoch:
                print(
                    "Validation loss has not improved for %d steps!"
                    % self.early_stopping_rounds
                )
                print("Early stopping applies.")
                break

        self.load_best_weights()
        return loss_history, val_loss_history

    def get_logits(self, x: torch.Tensor, with_grad: bool) -> torch.Tensor:
        if not with_grad:
            with torch.no_grad():
                # with_grad = True to avoid recursion
                return self.get_logits(x, with_grad=True)

        # Warning: use wrapper model
        preds = self.wrapper_model(x)
        if self.objective == "binary":
            preds = torch.sigmoid(preds)
        elif self.objective == "classification":
            preds = F.softmax(preds, dim=1)
        return preds

    def predict_helper(self, x, keep_grad=False):
        self.model.eval()

        x = np.array(x, dtype=float)
        x = torch.tensor(x).float()

        test_dataset = TensorDataset(x)
        test_loader = DataLoader(
            dataset=test_dataset,
            batch_size=self.val_batch_size,
            shuffle=False,
            num_workers=2,
        )

        predictions = []
        for batch_x in test_loader:
            preds = self.get_logits(batch_x[0], with_grad=keep_grad)
            predictions.append(preds.cpu())

        return np.concatenate(predictions)

    def save(self, path: str) -> None:
        kwargs_save = deepcopy(self.constructor_kwargs)
        if "x_metadata" in kwargs_save:
            kwargs_save.pop("x_metadata")

        args_path = f"{path}/args.json"
        save_json(kwargs_save, parent_exists(args_path))

        weigths_path = f"{path}/weights.pt"
        torch.save(self.model.state_dict(), parent_exists(weigths_path))

    @classmethod
    def load_class(cls, path: str, **kwargs: Dict[str, Any]) -> BaseModelTorch:

        args_path = f"{path}/args.json"
        weigths_path = f"{path}/weights.pt"

        load_kwargs = load_json(args_path)
        model = cls(**load_kwargs, **kwargs)
        model.model.load_state_dict(torch.load(weigths_path))
        model.model.to(model.device)

        return model

    def attribute(self, X, y, strategy=""):
        """Generate feature attributions for the model input.
        Two strategies are supported: default ("") or "diag".
        The default strategie takes the sum
        over a column of the attention map, while "diag"
        returns only the diagonal (feature attention to itself)
        of the attention map.
        return array with the same shape as X.
        """
        global my_attention
        # self.load_model(filename_extension="best", directory="tmp")

        X = {"data": X, "mask": np.ones_like(X)}
        y = {"data": np.ones((X["data"].shape[0], 1))}

        test_ds = DataSetCatCon(X, y, self.cat_idx, self.objective)
        testloader = DataLoader(
            test_ds,
            batch_size=self.val_batch_size,
            shuffle=False,
            num_workers=4,
        )

        self.model.eval()
        # print(self.model)
        # Apply hook.
        my_attention = torch.zeros(0)

        def sample_attribution(layer, minput, output):
            global my_attention
            # print(minput)
            """ an hook to extract the attention maps. """
            h = layer.heads
            q, k, v = layer.to_qkv(minput[0]).chunk(3, dim=-1)
            q, k, v = map(
                lambda t: rearrange(t, "b n (h d) -> b h n d", h=h), (q, k, v)
            )
            sim = einsum("b h i d, b h j d -> b h i j", q, k) * layer.scale
            my_attention = sim.softmax(dim=-1)

        # print(type(self.model.transformer.layers[0][0].fn.fn))
        self.model.transformer.layers[0][0].fn.fn.register_forward_hook(
            sample_attribution
        )
        attributions = []
        with torch.no_grad():
            for data in testloader:
                x_categ, x_cont, y_gts, cat_mask, con_mask = data

                x_categ, x_cont = x_categ.to(self.device), x_cont.to(
                    self.device
                )
                cat_mask, con_mask = cat_mask.to(self.device), con_mask.to(
                    self.device
                )
                # print(x_categ.shape, x_cont.shape)
                _, x_categ_enc, x_cont_enc = embed_data_mask(
                    x_categ, x_cont, cat_mask, con_mask, self.model
                )
                reps = self.model.transformer(x_categ_enc, x_cont_enc)
                # y_reps = reps[:, 0, :]
                # y_outs = self.model.mlpfory(y_reps)
                if strategy == "diag":
                    attributions.append(
                        my_attention.sum(dim=1)[:, 1:, 1:].diagonal(0, 1, 2)
                    )
                else:
                    attributions.append(
                        my_attention.sum(dim=1)[:, 1:, 1:].sum(dim=1)
                    )

        attributions = np.concatenate(attributions)
        return attributions

    @staticmethod
    def define_trial_parameters(trial: optuna.Trial, trial_params: Dict[str, Any]):
        params = {
            "dim": trial.suggest_categorical("dim", [32, 64, 128, 256]),
            "depth": trial.suggest_categorical("depth", [1, 2, 3, 6, 12]),
            "heads": trial.suggest_categorical("heads", [2, 4, 8]),
            "dropout": trial.suggest_categorical(
                "dropout", [0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]
            ),
        }
        return params


models: List[Tuple[str, Type[Model]]] = [("saint", SAINT)]



from mlc.logging.comet_config import COMET_APIKEY, COMET_PROJECT, COMET_WORKSPACE
from comet_ml import Experiment
import time

class XP(object):
    def __init__(self,args, project_name=COMET_PROJECT, workspace=COMET_WORKSPACE):

        if project_name == "" or project_name == None:
            self.xp = None
        else:
            timestamp = time.time()
            args["timestamp"] = timestamp
            experiment_name = "mcl_{}_{}_{}_{}".format(args.get("model_name", ""), args.get("dataset", ""),
                                                           args.get("trial", ""),timestamp)
            experiment = Experiment(api_key=COMET_APIKEY,
                                    project_name=project_name,
                                    workspace=workspace,
                                    auto_param_logging=False, auto_metric_logging=False,
                                    parse_args=False, display_summary=False, disabled=False)

            experiment.set_name(experiment_name)
            experiment.log_parameters(args)

            self.xp = experiment

    def log_parameters(self, *args, **kwargs):
        if self.xp is not None:
            self.xp.log_parameters(*args,**kwargs)
        else:
            print("logging parameters", *args)
            print("logging parameters", kwargs)


    def log_metrics(self, *args, **kwargs):
        if self.xp is not None:
            self.xp.log_metrics(*args,**kwargs)
        else:
            print("logging metrics", *args)
            print("logging metrics", kwargs)

    def log_metric(self, name, value, **kwargs):
        if self.xp is not None:
            self.xp.log_metric(name, value,**kwargs)
        else:
            print("logging metric", name, value)
            print("logging metric", kwargs)

    def end(self):
        if self.xp is not None:
            self.xp.end()

    def log_model(self, name,path):
        if self.xp is not None:
            self.xp.log_model(name, path, copy_to_tmp=False)
        else:
            print("Logging model", name, path)

    def get_name(self):
        if self.xp is not None:
            return self.xp.get_name()+"_"
        else:
            return ""




"""
Based on https://github.com/openai/guided-diffusion/blob/main/guided_diffusion
and https://github.com/ehoogeboom/multinomial_diffusion
"""

import math

from .utils import *

"""
Based in part on: https://github.com/lucidrains/denoising-diffusion-pytorch/blob/5989f4c77eafcdc6be0fb4739f0f277a6dd7f7d8/denoising_diffusion_pytorch/denoising_diffusion_pytorch.py#L281
"""
eps = 1e-8


def get_named_beta_schedule(schedule_name, num_diffusion_timesteps):
    """
    Get a pre-defined beta schedule for the given name.
    The beta schedule library consists of beta schedules which remain similar
    in the limit of num_diffusion_timesteps.
    Beta schedules may be added, but should not be removed or changed once
    they are committed to maintain backwards compatibility.
    """
    if schedule_name == "linear":
        # Linear schedule from Ho et al, extended to work for any number of
        # diffusion steps.
        scale = 1000 / num_diffusion_timesteps
        beta_start = scale * 0.0001
        beta_end = scale * 0.02
        return np.linspace(
            beta_start, beta_end, num_diffusion_timesteps, dtype=np.float64
        )
    elif schedule_name == "cosine":
        return betas_for_alpha_bar(
            num_diffusion_timesteps,
            lambda t: math.cos((t + 0.008) / 1.008 * math.pi / 2) ** 2,
        )
    else:
        raise NotImplementedError(f"unknown beta schedule: {schedule_name}")


def betas_for_alpha_bar(num_diffusion_timesteps, alpha_bar, max_beta=0.999):
    """
    Create a beta schedule that discretizes the given alpha_t_bar function,
    which defines the cumulative product of (1-beta) over time from t = [0,1].
    :param num_diffusion_timesteps: the number of betas to produce.
    :param alpha_bar: a lambda that takes an argument t from 0 to 1 and
                      produces the cumulative product of (1-beta) up to that
                      part of the diffusion process.
    :param max_beta: the maximum beta to use; use values lower than 1 to
                     prevent singularities.
    """
    betas = []
    for i in range(num_diffusion_timesteps):
        t1 = i / num_diffusion_timesteps
        t2 = (i + 1) / num_diffusion_timesteps
        betas.append(min(1 - alpha_bar(t2) / alpha_bar(t1), max_beta))
    return np.array(betas)


class GaussianMultinomialDiffusion(torch.nn.Module):
    def __init__(
            self,
            num_classes: np.array,
            num_numerical_features: int,
            denoise_fn,
            num_timesteps=1000,
            gaussian_loss_type='mse',
            gaussian_parametrization='eps',
            multinomial_loss_type='vb_stochastic',
            parametrization='x0',
            scheduler='cosine',
            device=torch.device('cpu')
    ):

        super(GaussianMultinomialDiffusion, self).__init__()
        assert multinomial_loss_type in ('vb_stochastic', 'vb_all')
        assert parametrization in ('x0', 'direct')

        if multinomial_loss_type == 'vb_all':
            print('Computing the loss using the bound on _all_ timesteps.'
                  ' This is expensive both in terms of memory and computation.')

        self.num_numerical_features = num_numerical_features
        self.num_classes = num_classes  # it as a vector [K1, K2, ..., Km]
        self.num_classes_expanded = torch.from_numpy(
            np.concatenate([num_classes[i].repeat(num_classes[i]) for i in range(len(num_classes))])
        ).to(device)

        self.slices_for_classes = [np.arange(self.num_classes[0])]
        offsets = np.cumsum(self.num_classes)
        for i in range(1, len(offsets)):
            self.slices_for_classes.append(np.arange(offsets[i - 1], offsets[i]))
        self.offsets = torch.from_numpy(np.append([0], offsets)).to(device)

        self._denoise_fn = denoise_fn
        self.gaussian_loss_type = gaussian_loss_type
        self.gaussian_parametrization = gaussian_parametrization
        self.multinomial_loss_type = multinomial_loss_type
        self.num_timesteps = num_timesteps
        self.parametrization = parametrization
        self.scheduler = scheduler

        alphas = 1. - get_named_beta_schedule(scheduler, num_timesteps)
        alphas = torch.tensor(alphas.astype('float64'))
        betas = 1. - alphas

        log_alpha = np.log(alphas)
        log_cumprod_alpha = np.cumsum(log_alpha)

        log_1_min_alpha = log_1_min_a(log_alpha)
        log_1_min_cumprod_alpha = log_1_min_a(log_cumprod_alpha)

        alphas_cumprod = np.cumprod(alphas, axis=0)
        alphas_cumprod_prev = torch.tensor(np.append(1.0, alphas_cumprod[:-1]))
        alphas_cumprod_next = torch.tensor(np.append(alphas_cumprod[1:], 0.0))
        sqrt_alphas_cumprod = np.sqrt(alphas_cumprod)
        sqrt_one_minus_alphas_cumprod = np.sqrt(1.0 - alphas_cumprod)
        sqrt_recip_alphas_cumprod = np.sqrt(1.0 / alphas_cumprod)
        sqrt_recipm1_alphas_cumprod = np.sqrt(1.0 / alphas_cumprod - 1)

        # Gaussian diffusion

        self.posterior_variance = (
                betas * (1.0 - alphas_cumprod_prev) / (1.0 - alphas_cumprod)
        )
        self.posterior_log_variance_clipped = torch.from_numpy(
            np.log(np.append(self.posterior_variance[1], self.posterior_variance[1:]))
        ).float().to(device)
        self.posterior_mean_coef1 = (
                betas * np.sqrt(alphas_cumprod_prev) / (1.0 - alphas_cumprod)
        ).float().to(device)
        self.posterior_mean_coef2 = (
                (1.0 - alphas_cumprod_prev)
                * np.sqrt(alphas.numpy())
                / (1.0 - alphas_cumprod)
        ).float().to(device)

        assert log_add_exp(log_alpha, log_1_min_alpha).abs().sum().item() < 1.e-5
        assert log_add_exp(log_cumprod_alpha, log_1_min_cumprod_alpha).abs().sum().item() < 1e-5
        assert (np.cumsum(log_alpha) - log_cumprod_alpha).abs().sum().item() < 1.e-5

        # Convert to float32 and register buffers.
        self.register_buffer('alphas', alphas.float().to(device))
        self.register_buffer('log_alpha', log_alpha.float().to(device))
        self.register_buffer('log_1_min_alpha', log_1_min_alpha.float().to(device))
        self.register_buffer('log_1_min_cumprod_alpha', log_1_min_cumprod_alpha.float().to(device))
        self.register_buffer('log_cumprod_alpha', log_cumprod_alpha.float().to(device))
        self.register_buffer('alphas_cumprod', alphas_cumprod.float().to(device))
        self.register_buffer('alphas_cumprod_prev', alphas_cumprod_prev.float().to(device))
        self.register_buffer('alphas_cumprod_next', alphas_cumprod_next.float().to(device))
        self.register_buffer('sqrt_alphas_cumprod', sqrt_alphas_cumprod.float().to(device))
        self.register_buffer('sqrt_one_minus_alphas_cumprod', sqrt_one_minus_alphas_cumprod.float().to(device))
        self.register_buffer('sqrt_recip_alphas_cumprod', sqrt_recip_alphas_cumprod.float().to(device))
        self.register_buffer('sqrt_recipm1_alphas_cumprod', sqrt_recipm1_alphas_cumprod.float().to(device))

        self.register_buffer('Lt_history', torch.zeros(num_timesteps))
        self.register_buffer('Lt_count', torch.zeros(num_timesteps))

    # Gaussian part
    def gaussian_q_mean_variance(self, x_start, t):
        mean = (
                extract(self.sqrt_alphas_cumprod, t, x_start.shape) * x_start
        )
        variance = extract(1.0 - self.alphas_cumprod, t, x_start.shape)
        log_variance = extract(
            self.log_1_min_cumprod_alpha, t, x_start.shape
        )
        return mean, variance, log_variance

    def gaussian_q_sample(self, x_start, t, noise=None):
        if noise is None:
            noise = torch.randn_like(x_start)
        assert noise.shape == x_start.shape
        return (
                extract(self.sqrt_alphas_cumprod, t, x_start.shape) * x_start
                + extract(self.sqrt_one_minus_alphas_cumprod, t, x_start.shape)
                * noise
        )

    def gaussian_q_posterior_mean_variance(self, x_start, x_t, t):
        assert x_start.shape == x_t.shape
        posterior_mean = (
                extract(self.posterior_mean_coef1, t, x_t.shape) * x_start
                + extract(self.posterior_mean_coef2, t, x_t.shape) * x_t
        )
        posterior_variance = extract(self.posterior_variance, t, x_t.shape)
        posterior_log_variance_clipped = extract(
            self.posterior_log_variance_clipped, t, x_t.shape
        )
        assert (
                posterior_mean.shape[0]
                == posterior_variance.shape[0]
                == posterior_log_variance_clipped.shape[0]
                == x_start.shape[0]
        )
        return posterior_mean, posterior_variance, posterior_log_variance_clipped

    def gaussian_p_mean_variance(
            self, model_output, x, t, clip_denoised=False, denoised_fn=None, model_kwargs=None
    ):
        if model_kwargs is None:
            model_kwargs = {}

        B, C = x.shape[:2]
        assert t.shape == (B,)

        model_variance = torch.cat([self.posterior_variance[1].unsqueeze(0).to(x.device), (1. - self.alphas)[1:]],
                                   dim=0)
        # model_variance = self.posterior_variance.to(x.device)
        model_log_variance = torch.log(model_variance)

        model_variance = extract(model_variance, t, x.shape)
        model_log_variance = extract(model_log_variance, t, x.shape)

        if self.gaussian_parametrization == 'eps':
            pred_xstart = self._predict_xstart_from_eps(x_t=x, t=t, eps=model_output)
        elif self.gaussian_parametrization == 'x0':
            pred_xstart = model_output
        else:
            raise NotImplementedError

        model_mean, _, _ = self.gaussian_q_posterior_mean_variance(
            x_start=pred_xstart, x_t=x, t=t
        )

        assert (
                model_mean.shape == model_log_variance.shape == pred_xstart.shape == x.shape
        ), f'{model_mean.shape}, {model_log_variance.shape}, {pred_xstart.shape}, {x.shape}'

        return {
            "mean": model_mean,
            "variance": model_variance,
            "log_variance": model_log_variance,
            "pred_xstart": pred_xstart,
        }

    def _vb_terms_bpd(
            self, model_output, x_start, x_t, t, clip_denoised=False, model_kwargs=None
    ):
        true_mean, _, true_log_variance_clipped = self.gaussian_q_posterior_mean_variance(
            x_start=x_start, x_t=x_t, t=t
        )
        out = self.gaussian_p_mean_variance(
            model_output, x_t, t, clip_denoised=clip_denoised, model_kwargs=model_kwargs
        )
        kl = normal_kl(
            true_mean, true_log_variance_clipped, out["mean"], out["log_variance"]
        )
        kl = mean_flat(kl) / np.log(2.0)

        decoder_nll = -discretized_gaussian_log_likelihood(
            x_start, means=out["mean"], log_scales=0.5 * out["log_variance"]
        )
        assert decoder_nll.shape == x_start.shape
        decoder_nll = mean_flat(decoder_nll) / np.log(2.0)

        # At the first timestep return the decoder NLL,
        # otherwise return KL(q(x_{t-1}|x_t,x_0) || p(x_{t-1}|x_t))
        output = torch.where((t == 0), decoder_nll, kl)
        return {"output": output, "pred_xstart": out["pred_xstart"], "out_mean": out["mean"], "true_mean": true_mean}

    def _prior_gaussian(self, x_start):
        """
        Get the prior KL term for the variational lower-bound, measured in
        bits-per-dim.

        This term can't be optimized, as it only depends on the encoder.

        :param x_start: the [N x C x ...] tensor of inputs.
        :return: a batch of [N] KL values (in bits), one per batch element.
        """
        batch_size = x_start.shape[0]
        t = torch.tensor([self.num_timesteps - 1] * batch_size, device=x_start.device)
        qt_mean, _, qt_log_variance = self.gaussian_q_mean_variance(x_start, t)
        kl_prior = normal_kl(
            mean1=qt_mean, logvar1=qt_log_variance, mean2=0.0, logvar2=0.0
        )
        return mean_flat(kl_prior) / np.log(2.0)

    def _gaussian_loss(self, model_out, x_start, x_t, t, noise, model_kwargs=None):
        if model_kwargs is None:
            model_kwargs = {}

        terms = {}
        if self.gaussian_loss_type == 'mse':
            terms["loss"] = mean_flat((noise - model_out) ** 2)
        elif self.gaussian_loss_type == 'kl':
            terms["loss"] = self._vb_terms_bpd(
                model_output=model_out,
                x_start=x_start,
                x_t=x_t,
                t=t,
                clip_denoised=False,
                model_kwargs=model_kwargs,
            )["output"]

        return terms['loss']

    def _predict_xstart_from_eps(self, x_t, t, eps):
        assert x_t.shape == eps.shape
        return (
                extract(self.sqrt_recip_alphas_cumprod, t, x_t.shape) * x_t
                - extract(self.sqrt_recipm1_alphas_cumprod, t, x_t.shape) * eps
        )

    def _predict_eps_from_xstart(self, x_t, t, pred_xstart):
        return (
                       extract(self.sqrt_recip_alphas_cumprod, t, x_t.shape) * x_t
                       - pred_xstart
               ) / extract(self.sqrt_recipm1_alphas_cumprod, t, x_t.shape)

    def gaussian_p_sample(
            self,
            model_out,
            x,
            t,
            clip_denoised=False,
            denoised_fn=None,
            model_kwargs=None,
    ):
        out = self.gaussian_p_mean_variance(
            model_out,
            x,
            t,
            clip_denoised=clip_denoised,
            denoised_fn=denoised_fn,
            model_kwargs=model_kwargs,
        )
        noise = torch.randn_like(x)
        nonzero_mask = (
            (t != 0).float().view(-1, *([1] * (len(x.shape) - 1)))
        )  # no noise when t == 0

        sample = out["mean"] + nonzero_mask * torch.exp(0.5 * out["log_variance"]) * noise
        return {"sample": sample, "pred_xstart": out["pred_xstart"]}

    # Multinomial part

    def multinomial_kl(self, log_prob1, log_prob2):
        kl = (log_prob1.exp() * (log_prob1 - log_prob2)).sum(dim=1)
        return kl

    def q_pred_one_timestep(self, log_x_t, t):
        log_alpha_t = extract(self.log_alpha, t, log_x_t.shape)
        log_1_min_alpha_t = extract(self.log_1_min_alpha, t, log_x_t.shape)

        # alpha_t * E[xt] + (1 - alpha_t) 1 / K
        log_probs = log_add_exp(
            log_x_t + log_alpha_t,
            log_1_min_alpha_t - torch.log(self.num_classes_expanded)
        )

        return log_probs

    def q_pred(self, log_x_start, t):
        log_cumprod_alpha_t = extract(self.log_cumprod_alpha, t, log_x_start.shape)
        log_1_min_cumprod_alpha = extract(self.log_1_min_cumprod_alpha, t, log_x_start.shape)

        log_probs = log_add_exp(
            log_x_start + log_cumprod_alpha_t,
            log_1_min_cumprod_alpha - torch.log(self.num_classes_expanded)
        )

        return log_probs

    def predict_start(self, model_out, log_x_t, t, out_dict):

        # model_out = self._denoise_fn(x_t, t.to(x_t.device), **out_dict)

        assert model_out.size(0) == log_x_t.size(0)
        assert model_out.size(1) == self.num_classes.sum(), f'{model_out.size()}'

        log_pred = torch.empty_like(model_out)
        for ix in self.slices_for_classes:
            log_pred[:, ix] = F.log_softmax(model_out[:, ix], dim=1)
        return log_pred

    def q_posterior(self, log_x_start, log_x_t, t):
        # q(xt-1 | xt, x0) = q(xt | xt-1, x0) * q(xt-1 | x0) / q(xt | x0)
        # where q(xt | xt-1, x0) = q(xt | xt-1).

        # EV_log_qxt_x0 = self.q_pred(log_x_start, t)

        # print('sum exp', EV_log_qxt_x0.exp().sum(1).mean())
        # assert False

        # log_qxt_x0 = (log_x_t.exp() * EV_log_qxt_x0).sum(dim=1)
        t_minus_1 = t - 1
        # Remove negative values, will not be used anyway for final decoder
        t_minus_1 = torch.where(t_minus_1 < 0, torch.zeros_like(t_minus_1), t_minus_1)
        log_EV_qxtmin_x0 = self.q_pred(log_x_start, t_minus_1)

        num_axes = (1,) * (len(log_x_start.size()) - 1)
        t_broadcast = t.to(log_x_start.device).view(-1, *num_axes) * torch.ones_like(log_x_start)
        log_EV_qxtmin_x0 = torch.where(t_broadcast == 0, log_x_start, log_EV_qxtmin_x0.to(torch.float32))

        # unnormed_logprobs = log_EV_qxtmin_x0 +
        #                     log q_pred_one_timestep(x_t, t)
        # Note: _NOT_ x_tmin1, which is how the formula is typically used!!!
        # Not very easy to see why this is true. But it is :)
        unnormed_logprobs = log_EV_qxtmin_x0 + self.q_pred_one_timestep(log_x_t, t)

        log_EV_xtmin_given_xt_given_xstart = \
            unnormed_logprobs \
            - sliced_logsumexp(unnormed_logprobs, self.offsets)

        return log_EV_xtmin_given_xt_given_xstart

    def p_pred(self, model_out, log_x, t, out_dict):
        if self.parametrization == 'x0':
            log_x_recon = self.predict_start(model_out, log_x, t=t, out_dict=out_dict)
            log_model_pred = self.q_posterior(
                log_x_start=log_x_recon, log_x_t=log_x, t=t)
        elif self.parametrization == 'direct':
            log_model_pred = self.predict_start(model_out, log_x, t=t, out_dict=out_dict)
        else:
            raise ValueError
        return log_model_pred

    @torch.no_grad()
    def p_sample(self, model_out, log_x, t, out_dict):
        model_log_prob = self.p_pred(model_out, log_x=log_x, t=t, out_dict=out_dict)
        out = self.log_sample_categorical(model_log_prob)
        return out

    @torch.no_grad()
    def p_sample_loop(self, shape, out_dict):
        device = self.log_alpha.device

        b = shape[0]
        # start with random normal image.
        img = torch.randn(shape, device=device)

        for i in reversed(range(1, self.num_timesteps)):
            img = self.p_sample(img, torch.full((b,), i, device=device, dtype=torch.long), out_dict)
        return img

    @torch.no_grad()
    def _sample(self, image_size, out_dict, batch_size=16):
        return self.p_sample_loop((batch_size, 3, image_size, image_size), out_dict)

    @torch.no_grad()
    def interpolate(self, x1, x2, t=None, lam=0.5):
        b, *_, device = *x1.shape, x1.device
        t = default(t, self.num_timesteps - 1)

        assert x1.shape == x2.shape

        t_batched = torch.stack([torch.tensor(t, device=device)] * b)
        xt1, xt2 = map(lambda x: self.q_sample(x, t=t_batched), (x1, x2))

        img = (1 - lam) * xt1 + lam * xt2
        for i in reversed(range(0, t)):
            img = self.p_sample(img, torch.full((b,), i, device=device, dtype=torch.long))

        return img

    def log_sample_categorical(self, logits):
        full_sample = []
        for i in range(len(self.num_classes)):
            one_class_logits = logits[:, self.slices_for_classes[i]]
            uniform = torch.rand_like(one_class_logits)
            gumbel_noise = -torch.log(-torch.log(uniform + 1e-30) + 1e-30)
            sample = (gumbel_noise + one_class_logits).argmax(dim=1)
            full_sample.append(sample.unsqueeze(1))
        full_sample = torch.cat(full_sample, dim=1)
        log_sample = index_to_log_onehot(full_sample, self.num_classes)
        return log_sample

    def q_sample(self, log_x_start, t):
        log_EV_qxt_x0 = self.q_pred(log_x_start, t)

        log_sample = self.log_sample_categorical(log_EV_qxt_x0)

        return log_sample

    def nll(self, log_x_start, out_dict):
        b = log_x_start.size(0)
        device = log_x_start.device
        loss = 0
        for t in range(0, self.num_timesteps):
            t_array = (torch.ones(b, device=device) * t).long()

            kl = self.compute_Lt(
                log_x_start=log_x_start,
                log_x_t=self.q_sample(log_x_start=log_x_start, t=t_array),
                t=t_array,
                out_dict=out_dict)

            loss += kl

        loss += self.kl_prior(log_x_start)

        return loss

    def kl_prior(self, log_x_start):
        b = log_x_start.size(0)
        device = log_x_start.device
        ones = torch.ones(b, device=device).long()

        log_qxT_prob = self.q_pred(log_x_start, t=(self.num_timesteps - 1) * ones)
        log_half_prob = -torch.log(self.num_classes_expanded * torch.ones_like(log_qxT_prob))

        kl_prior = self.multinomial_kl(log_qxT_prob, log_half_prob)
        return sum_except_batch(kl_prior)

    def compute_Lt(self, model_out, log_x_start, log_x_t, t, out_dict, detach_mean=False):
        log_true_prob = self.q_posterior(
            log_x_start=log_x_start, log_x_t=log_x_t, t=t)
        log_model_prob = self.p_pred(model_out, log_x=log_x_t, t=t, out_dict=out_dict)

        if detach_mean:
            log_model_prob = log_model_prob.detach()

        kl = self.multinomial_kl(log_true_prob, log_model_prob)
        kl = sum_except_batch(kl)

        decoder_nll = -log_categorical(log_x_start, log_model_prob)
        decoder_nll = sum_except_batch(decoder_nll)

        mask = (t == torch.zeros_like(t)).float()
        loss = mask * decoder_nll + (1. - mask) * kl

        return loss

    def sample_time(self, b, device, method='uniform'):
        if method == 'importance':
            if not (self.Lt_count > 10).all():
                return self.sample_time(b, device, method='uniform')

            Lt_sqrt = torch.sqrt(self.Lt_history + 1e-10) + 0.0001
            Lt_sqrt[0] = Lt_sqrt[1]  # Overwrite decoder term with L1.
            pt_all = (Lt_sqrt / Lt_sqrt.sum()).to(device)

            t = torch.multinomial(pt_all, num_samples=b, replacement=True).to(device)

            pt = pt_all.gather(dim=0, index=t)

            return t, pt

        elif method == 'uniform':
            t = torch.randint(0, self.num_timesteps, (b,), device=device).long()

            pt = torch.ones_like(t).float() / self.num_timesteps
            return t, pt
        else:
            raise ValueError

    def _multinomial_loss(self, model_out, log_x_start, log_x_t, t, pt, out_dict):

        if self.multinomial_loss_type == 'vb_stochastic':
            kl = self.compute_Lt(
                model_out, log_x_start, log_x_t, t, out_dict
            )
            kl_prior = self.kl_prior(log_x_start)
            # Upweigh loss term of the kl
            vb_loss = kl / pt + kl_prior

            return vb_loss

        elif self.multinomial_loss_type == 'vb_all':
            # Expensive, dont do it ;).
            # DEPRECATED
            return -self.nll(log_x_start)
        else:
            raise ValueError()

    def log_prob(self, x, out_dict):
        b, device = x.size(0), x.device
        if self.training:
            return self._multinomial_loss(x, out_dict)

        else:
            log_x_start = index_to_log_onehot(x, self.num_classes)

            t, pt = self.sample_time(b, device, 'importance')

            kl = self.compute_Lt(
                log_x_start, self.q_sample(log_x_start=log_x_start, t=t), t, out_dict)

            kl_prior = self.kl_prior(log_x_start)

            # Upweigh loss term of the kl
            loss = kl / pt + kl_prior

            return -loss

    def mixed_loss(self, x, out_dict):
        b = x.shape[0]
        device = x.device
        t, pt = self.sample_time(b, device, 'uniform')

        x_num = x[:, :self.num_numerical_features]
        x_cat = x[:, self.num_numerical_features:]

        x_num_t = x_num
        log_x_cat_t = x_cat
        if x_num.shape[1] > 0:
            noise = torch.randn_like(x_num)
            x_num_t = self.gaussian_q_sample(x_num, t, noise=noise)
        if x_cat.shape[1] > 0:
            log_x_cat = index_to_log_onehot(x_cat.long(), self.num_classes)
            log_x_cat_t = self.q_sample(log_x_start=log_x_cat, t=t)

        x_in = torch.cat([x_num_t, log_x_cat_t], dim=1)

        model_out = self._denoise_fn(
            x_in,
            t,
            **out_dict
        )

        model_out_num = model_out[:, :self.num_numerical_features]
        model_out_cat = model_out[:, self.num_numerical_features:]

        loss_multi = torch.zeros((1,)).float()
        loss_gauss = torch.zeros((1,)).float()
        if x_cat.shape[1] > 0:
            loss_multi = self._multinomial_loss(model_out_cat, log_x_cat, log_x_cat_t, t, pt, out_dict) / len(
                self.num_classes)

        if x_num.shape[1] > 0:
            loss_gauss = self._gaussian_loss(model_out_num, x_num, x_num_t, t, noise)

        # loss_multi = torch.where(out_dict['y'] == 1, loss_multi, 2 * loss_multi)
        # loss_gauss = torch.where(out_dict['y'] == 1, loss_gauss, 2 * loss_gauss)

        return loss_multi.mean(), loss_gauss.mean()

    @torch.no_grad()
    def mixed_elbo(self, x0, out_dict):
        b = x0.size(0)
        device = x0.device

        x_num = x0[:, :self.num_numerical_features]
        x_cat = x0[:, self.num_numerical_features:]
        has_cat = x_cat.shape[1] > 0
        if has_cat:
            log_x_cat = index_to_log_onehot(x_cat.long(), self.num_classes).to(device)

        gaussian_loss = []
        xstart_mse = []
        mse = []
        mu_mse = []
        out_mean = []
        true_mean = []
        multinomial_loss = []
        for t in range(self.num_timesteps):
            t_array = (torch.ones(b, device=device) * t).long()
            noise = torch.randn_like(x_num)

            x_num_t = self.gaussian_q_sample(x_start=x_num, t=t_array, noise=noise)
            if has_cat:
                log_x_cat_t = self.q_sample(log_x_start=log_x_cat, t=t_array)
            else:
                log_x_cat_t = x_cat

            model_out = self._denoise_fn(
                torch.cat([x_num_t, log_x_cat_t], dim=1),
                t_array,
                **out_dict
            )

            model_out_num = model_out[:, :self.num_numerical_features]
            model_out_cat = model_out[:, self.num_numerical_features:]

            kl = torch.tensor([0.0])
            if has_cat:
                kl = self.compute_Lt(
                    model_out=model_out_cat,
                    log_x_start=log_x_cat,
                    log_x_t=log_x_cat_t,
                    t=t_array,
                    out_dict=out_dict
                )

            out = self._vb_terms_bpd(
                model_out_num,
                x_start=x_num,
                x_t=x_num_t,
                t=t_array,
                clip_denoised=False
            )

            multinomial_loss.append(kl)
            gaussian_loss.append(out["output"])
            xstart_mse.append(mean_flat((out["pred_xstart"] - x_num) ** 2))
            # mu_mse.append(mean_flat(out["mean_mse"]))
            out_mean.append(mean_flat(out["out_mean"]))
            true_mean.append(mean_flat(out["true_mean"]))

            eps = self._predict_eps_from_xstart(x_num_t, t_array, out["pred_xstart"])
            mse.append(mean_flat((eps - noise) ** 2))

        gaussian_loss = torch.stack(gaussian_loss, dim=1)
        multinomial_loss = torch.stack(multinomial_loss, dim=1)
        xstart_mse = torch.stack(xstart_mse, dim=1)
        mse = torch.stack(mse, dim=1)
        # mu_mse = torch.stack(mu_mse, dim=1)
        out_mean = torch.stack(out_mean, dim=1)
        true_mean = torch.stack(true_mean, dim=1)

        prior_gauss = self._prior_gaussian(x_num)

        prior_multin = torch.tensor([0.0])
        if has_cat:
            prior_multin = self.kl_prior(log_x_cat)

        total_gauss = gaussian_loss.sum(dim=1) + prior_gauss
        total_multin = multinomial_loss.sum(dim=1) + prior_multin
        return {
            "total_gaussian": total_gauss,
            "total_multinomial": total_multin,
            "losses_gaussian": gaussian_loss,
            "losses_multinimial": multinomial_loss,
            "xstart_mse": xstart_mse,
            "mse": mse,
            # "mu_mse": mu_mse
            "out_mean": out_mean,
            "true_mean": true_mean
        }

    @torch.no_grad()
    def gaussian_ddim_step(
            self,
            model_out_num,
            x,
            t,
            clip_denoised=False,
            denoised_fn=None,
            eta=0.0
    ):
        out = self.gaussian_p_mean_variance(
            model_out_num,
            x,
            t,
            clip_denoised=clip_denoised,
            denoised_fn=denoised_fn,
            model_kwargs=None,
        )

        eps = self._predict_eps_from_xstart(x, t, out["pred_xstart"])

        alpha_bar = extract(self.alphas_cumprod, t, x.shape)
        alpha_bar_prev = extract(self.alphas_cumprod_prev, t, x.shape)
        sigma = (
                eta
                * torch.sqrt((1 - alpha_bar_prev) / (1 - alpha_bar))
                * torch.sqrt(1 - alpha_bar / alpha_bar_prev)
        )

        noise = torch.randn_like(x)
        mean_pred = (
                out["pred_xstart"] * torch.sqrt(alpha_bar_prev)
                + torch.sqrt(1 - alpha_bar_prev - sigma ** 2) * eps
        )
        nonzero_mask = (
            (t != 0).float().view(-1, *([1] * (len(x.shape) - 1)))
        )  # no noise when t == 0
        sample = mean_pred + nonzero_mask * sigma * noise

        return sample

    @torch.no_grad()
    def gaussian_ddim_sample(
            self,
            noise,
            T,
            out_dict,
            eta=0.0
    ):
        x = noise
        b = x.shape[0]
        device = x.device
        for t in reversed(range(T)):
            print(f'Sample timestep {t:4d}', end='\r')
            t_array = (torch.ones(b, device=device) * t).long()
            out_num = self._denoise_fn(x, t_array, **out_dict)
            x = self.gaussian_ddim_step(
                out_num,
                x,
                t_array
            )
        print()
        return x

    @torch.no_grad()
    def gaussian_ddim_reverse_step(
            self,
            model_out_num,
            x,
            t,
            clip_denoised=False,
            eta=0.0
    ):
        assert eta == 0.0, "Eta must be zero."
        out = self.gaussian_p_mean_variance(
            model_out_num,
            x,
            t,
            clip_denoised=clip_denoised,
            denoised_fn=None,
            model_kwargs=None,
        )

        eps = (
                      extract(self.sqrt_recip_alphas_cumprod, t, x.shape) * x
                      - out["pred_xstart"]
              ) / extract(self.sqrt_recipm1_alphas_cumprod, t, x.shape)
        alpha_bar_next = extract(self.alphas_cumprod_next, t, x.shape)

        mean_pred = (
                out["pred_xstart"] * torch.sqrt(alpha_bar_next)
                + torch.sqrt(1 - alpha_bar_next) * eps
        )

        return mean_pred

    @torch.no_grad()
    def gaussian_ddim_reverse_sample(
            self,
            x,
            T,
            out_dict,
    ):
        b = x.shape[0]
        device = x.device
        for t in range(T):
            print(f'Reverse timestep {t:4d}', end='\r')
            t_array = (torch.ones(b, device=device) * t).long()
            out_num = self._denoise_fn(x, t_array, **out_dict)
            x = self.gaussian_ddim_reverse_step(
                out_num,
                x,
                t_array,
                eta=0.0
            )
        print()

        return x

    @torch.no_grad()
    def multinomial_ddim_step(
            self,
            model_out_cat,
            log_x_t,
            t,
            out_dict,
            eta=0.0
    ):
        # not ddim, essentially
        log_x0 = self.predict_start(model_out_cat, log_x_t=log_x_t, t=t, out_dict=out_dict)

        alpha_bar = extract(self.alphas_cumprod, t, log_x_t.shape)
        alpha_bar_prev = extract(self.alphas_cumprod_prev, t, log_x_t.shape)
        sigma = (
                eta
                * torch.sqrt((1 - alpha_bar_prev) / (1 - alpha_bar))
                * torch.sqrt(1 - alpha_bar / alpha_bar_prev)
        )

        coef1 = sigma
        coef2 = alpha_bar_prev - sigma * alpha_bar
        coef3 = 1 - coef1 - coef2

        log_ps = torch.stack([
            torch.log(coef1) + log_x_t,
            torch.log(coef2) + log_x0,
            torch.log(coef3) - torch.log(self.num_classes_expanded)
        ], dim=2)

        log_prob = torch.logsumexp(log_ps, dim=2)

        out = self.log_sample_categorical(log_prob)

        return out

    @torch.no_grad()
    def sample_ddim(self, num_samples, y_dist):
        b = num_samples
        device = self.log_alpha.device
        z_norm = torch.randn((b, self.num_numerical_features), device=device)

        has_cat = self.num_classes[0] != 0
        log_z = torch.zeros((b, 0), device=device).float()
        if has_cat:
            uniform_logits = torch.zeros((b, len(self.num_classes_expanded)), device=device)
            log_z = self.log_sample_categorical(uniform_logits)

        y = torch.multinomial(
            y_dist,
            num_samples=b,
            replacement=True
        )
        out_dict = {'y': y.long().to(device)}
        for i in reversed(range(0, self.num_timesteps)):
            print(f'Sample timestep {i:4d}', end='\r')
            t = torch.full((b,), i, device=device, dtype=torch.long)
            model_out = self._denoise_fn(
                torch.cat([z_norm, log_z], dim=1).float(),
                t,
                **out_dict
            )
            model_out_num = model_out[:, :self.num_numerical_features]
            model_out_cat = model_out[:, self.num_numerical_features:]
            z_norm = self.gaussian_ddim_step(model_out_num, z_norm, t, clip_denoised=False)
            if has_cat:
                log_z = self.multinomial_ddim_step(model_out_cat, log_z, t, out_dict)

        print()
        z_ohe = torch.exp(log_z).round()
        z_cat = log_z
        if has_cat:
            z_cat = ohe_to_categories(z_ohe, self.num_classes)
        sample = torch.cat([z_norm, z_cat], dim=1).cpu()
        return sample, out_dict

    @torch.no_grad()
    def sample(self, num_samples, y_dist):
        b = num_samples
        device = self.log_alpha.device
        z_norm = torch.randn((b, self.num_numerical_features), device=device)

        has_cat = self.num_classes[0] != 0
        log_z = torch.zeros((b, 0), device=device).float()
        if has_cat:
            uniform_logits = torch.zeros((b, len(self.num_classes_expanded)), device=device)
            log_z = self.log_sample_categorical(uniform_logits)

        y = torch.multinomial(
            y_dist,
            num_samples=b,
            replacement=True
        )
        out_dict = {'y': y.long().to(device)}
        for i in reversed(range(0, self.num_timesteps)):
            print(f'Sample timestep {i:4d}', end='\r')
            t = torch.full((b,), i, device=device, dtype=torch.long)
            model_out = self._denoise_fn(
                torch.cat([z_norm, log_z], dim=1).float(),
                t,
                **out_dict
            )
            model_out_num = model_out[:, :self.num_numerical_features]
            model_out_cat = model_out[:, self.num_numerical_features:]
            z_norm = self.gaussian_p_sample(model_out_num, z_norm, t, clip_denoised=False)['sample']
            if has_cat:
                log_z = self.p_sample(model_out_cat, log_z, t, out_dict)

        print()
        z_ohe = torch.exp(log_z).round()
        z_cat = log_z
        if has_cat:
            z_cat = ohe_to_categories(z_ohe, self.num_classes)
        sample = torch.cat([z_norm, z_cat], dim=1).cpu()
        return sample, out_dict

    def sample_all(self, num_samples, batch_size, y_dist, ddim=False):
        if ddim:
            print('Sample using DDIM.')
            sample_fn = self.sample_ddim
        else:
            sample_fn = self.sample

        b = batch_size

        all_y = []
        all_samples = []
        num_generated = 0
        while num_generated < num_samples:
            sample, out_dict = sample_fn(b, y_dist)
            mask_nan = torch.any(sample.isnan(), dim=1)
            sample = sample[~mask_nan]
            out_dict['y'] = out_dict['y'][~mask_nan]

            all_samples.append(sample)
            all_y.append(out_dict['y'].cpu())
            if sample.shape[0] != b:
                raise FoundNANsError
            num_generated += sample.shape[0]

        x_gen = torch.cat(all_samples, dim=0)[:num_samples]
        y_gen = torch.cat(all_y, dim=0)[:num_samples]

        return x_gen, y_gen



from typing import Union

import numpy as np
import numpy.typing as npt
import pandas as pd

from mlc.transformers.transformer import Transformer


class IdentityTransformer(Transformer):
    def fit(
        self,
        x: Union[npt.NDArray[np.float_], pd.DataFrame],
        y: Union[npt.NDArray[np.int_], pd.Series] = None,
    ) -> None:
        # Fit the interface
        pass

    def transform(
        self, x: Union[npt.NDArray[np.float_], pd.DataFrame]
    ) -> Union[npt.NDArray[np.float_], pd.DataFrame]:
        return x

    def load(self, path: str) -> None:
        # Nothing to load
        pass

    def save(self, path: str) -> None:
        # Nothing to save
        pass
